/*
 * beam.c
 * code to beam texts across the TCP connection following a special protocol
 * source file for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>

#include "cancan.h"
#include "term.h"
#include "parse.h"
#include "beam.h"

int pause();
int gettimeofday();

static int next_digit();
static void edit_message();
static void finish_edit();

editsess *edit_sess;		/* head of session list */

char *edit_start, *edit_end;	/* when starting/leaving editing sessions */

/*
 * read in another decimal. return 0 in case of overflow or invalid char.
 */
static int next_digit(val, cp)
unsigned long *val;
char *cp;
{
    unsigned int i;
    return (i = ctoi(*cp)) < 10 ? (*val = *val * 10 + i, 1) : 0;
}

/*
 * Process editing protocol message with len chars.
 * Return number of characters used in the message.
 */
int mp_receive(sesp, data, len)
editsess **sesp;
char *data;
int len;
{
    editsess *s = *sesp;
    char *src = data;
    int state, l;
    if(!s) {
	*sesp = s = (editsess*)calloc(1, sizeof(editsess));
	if(!s) syserr("calloc");
	/* NOTE: a null pointer isn't necessarily all-bits-zero? */
	/* s->next = s->fname = 0; */
	if(edit_sess) {
	    s->next = edit_sess;
	    s->num = edit_sess->num;
	}
	s->num++;  edit_sess = s;
	src += strlen(MPI);  len -= strlen(MPI);
    }
    if((state = s->state) > 1 && len > (l = s->size - s->pos)) len = l;
    l = len;
    while(l) {
	switch(state) {
	case 0:					/* 'E' or 'V'	*/
		if(*src != 'E' && *src != 'V') break;
		s->mode = *src++;  ++state;  if(!--l) break;
		/* FALLTHROUGH */
	case 1:					/* message size	*/
		while(next_digit(&s->size, src) && (++src, --l));
		if(!l || !s->size) break;  ++state;
		if(l > s->size) l = s->size;
		continue;
	case 4:					/* 'M' if edit	*/
		len = l;	/* message body starts */
		if(s->mode != 'E') { state = 8;  continue; }
		if(*src != 'M') break;
		++src;  ++state;  if(!--l) break;
		/* FALLTHROUGH */
	case 5:					/* message key	*/
		while(next_digit(&s->key, src) && (++src, --l));
		if(!l || !s->key) break;  ++state;  continue;
	case 8:					/* description	*/
		/* keep description length in s->fd; it's not used yet */
		for( ; l && isprint(*src); ++src, --l)
		    if(s->fd < MAXDESCLEN - 1) s->desc[s->fd++] = *src;
		if(!l) break;
		fputs(s->desc, stdout);  newline();
		++state;
		/* FALLTHROUGH */
	case 2: case 6: case 9:			/* CR's		*/
		++state;  if(*src == '\r' && (++src, !--l)) break;
		/* FALLTHROUGH */
	case 3: case 7: case 10:		/* LF's		*/
		if(*src != '\n') break;
		++state;  ++src;  --l;
		if(state != 11) continue;
		if(s->mode) {			/* create file	*/
		    char mbuf[32];
		    struct timeval tt;
		    sprintf(mbuf, "/tmp/%c%lu", s->mode,
			    s->key ? s->key : (unsigned long)
			    (gettimeofday(&tt, (struct timezone *)0), tt.tv_sec));
		    if((s->fd = open(mbuf, O_WRONLY | O_CREAT, 0666)) < 0) {
			syswarn("#could not create temp file");
			s->mode = 0;
		    } else
			s->fname = str_dup(mbuf);
		}
		if(!l) break;
		/* FALLTHROUGH */
	case 11:				/* write file	*/
		footnote[0] = 0;
		if(!s->mode) {
		    /*
		     * FEATURE: a short packet is assumed to mean
		     * that the data transmission was aborted.
		     */
		    /* if(len < 256 && len < s->size - s->pos) break; */
		    if(len < 256) break;
		} else if(write(s->fd, src, l) < l) {
		    syswarn("#could not write temp file");
		    s->mode = 0;
		}
		src += l;  l = 0;
		if(s->pos + len < s->size)
		    sprintf(footnote, "read %lu bytes out of %lu",
			    s->pos + len, s->size);
	} /* end switch */
	break;
    } /* end while */
    if(((s->state = state) > 4 && (s->pos += len - l) >= s->size) || l) {
	*sesp = 0;
	if(l || state < 11) {
	    if(s->mode)
		strcpy(footnote, "#warning: mume protocol error");
	    finish_edit(s);
	} else
	    edit_message(s);
    }
    return src - data;
}    

static void edit_message(s)
editsess *s;
{
    char *editor;
    int waitforeditor = 0;
    if(s->mode == 'E') {
	struct stat sbuf;
	if(fstat(s->fd, &sbuf) < 0) syserr("fstat");	/* file is open */
	s->size = sbuf.st_size;
	s->time = sbuf.st_mtime;
	if(!(editor = getenv("CANCANEDITOR")) && !(editor = getenv("EDITOR")))
	    editor = DEFAULT_EDITOR;
    } else if(s->mode == 'V') {
	if(!(editor = getenv("CANCANPAGER")) && !(editor = getenv("PAGER")))
	    editor = DEFAULT_PAGER;
    } else {
	finish_edit(s);
	return;
    }
    close(s->fd);
    /* send an edit_start message (if wanted) */
    if(edit_start && !edit_sess->next)
	parse(edit_start, 0, (struct param *)0);
    if(editor[0] == '&')
	editor++;
    else {
	waitforeditor = 1;
	reset_terminal();
    }
    fflush(stdout);
    waiting_for_pid = fork();
    if(!waiting_for_pid) {
	char *args[] = { "/bin/sh", "-c", inserted_next, 0 };
	sprintf(inserted_next, "%s %s", editor, s->fname);
	if(!waitforeditor) { close(0);  close(1); }
	set_env("TITLE", s->desc);
	execve(args[0], args, environ);
	_exit(1);
    } else if(waiting_for_pid < 0) {
	printf("#fork failed, sorry\n");
	s->mode = 0;
	finish_edit(s);
	if(!waitforeditor) return;
    } else {
	s->pid = waiting_for_pid;
	if(!waitforeditor) return;
	flush_output_buffer();
	/* childsig() will call finish_edit */
	while(waiting_for_pid) pause();
	lastline();
    }
    set_terminal();
}

static void finish_edit(s)
editsess *s;
{
    union {
	char buf[BUFSIZE];
	struct stat st;
    } b;
    editsess **sp;
    unsigned long size;
    int i, back_data_sent = data_sent;
#ifndef OMIT_SESSIONS
    int back_fd = sockfd;
    sockfd = main_sess(); 
#endif
    if(s->mode == 'E' && s->fname)
	if((s->fd = open(s->fname, O_RDONLY)) < 0 || fstat(s->fd, &b.st) < 0 ||
		((size = b.st.st_size) == s->size && b.st.st_mtime == s->time))
	    s->mode = 0;
	else {
	    /* file was changed by editor: send back result to server */
	    send_to_host(b.buf, sprintf(b.buf, "%sE%lu\nE%lu\n", MPI,
		    sprintf(b.buf, "_%lu_", s->key) + size, s->key));
	    for( ; size; size -= i) {
		i = sizeof(b.buf) < size ? sizeof(b.buf) : size;
		if((i = read(s->fd, b.buf, i)) > 0)
		    send_to_host(b.buf, i);
		else
		    syserr("read temp");
	    }
	    sprintf(footnote, "#session %d completed", s->num);
	}
    if(s->mode != 'E' && s->key) {		/* cancel editing session */
	send_to_host(b.buf, sprintf(b.buf, "%sE%lu\nC%lu\n", MPI,
		(unsigned long)sprintf(b.buf, "_%lu_", s->key), s->key));
	sprintf(footnote, "#session %d cancelled", s->num);
    }
    data_sent = back_data_sent;		/* MP messages aren't prompted	*/
    if(s->fname) {
	close(s->fd);
	unlink(s->fname);
	free(s->fname);
    }
    for(sp = &edit_sess; *sp; sp = &(*sp)->next)
	if(*sp == s) { *sp = s->next;  break; }
    free((char*)s);
    /* parse (send) the edit_end message if this is the last editor... */
    if(edit_end && !edit_sess)
	parse(edit_end, 0, (struct param *)0);
#ifndef OMIT_SESSIONS
    sockfd = back_fd;
#endif
}

/*
 * cancel an editing session; does not free anything
 * (the child death signal handler will do that)
 */
void cancel_edit(s)
editsess *s;
{
    int sig = s->mode ? SIGTERM : SIGKILL;
    s->mode = 0;
    if(s->pid)	/* Editicide */
	kill(s->pid, sig);
}

/*
 * Our child has snuffed it.
 * If it was an editor/viewer, finish the session.
 */
void childsig()
{
    static int result;
    editsess *sp;
    int pid;
    /* NOTE: should we kick awake sleepy editors? */
    while((pid = waitpid(-1, &result, WNOHANG)) > 0) {
	if(pid == waiting_for_pid) {
	    waiting_for_pid = 0;
	    wait_result = result;
	}
	for(sp = edit_sess; sp; sp = sp->next)
	    if(sp->pid == pid) {
		if(result) sp->mode = 0;
		finish_edit(sp);	/* NOTE: we do a lot of work here... */
		break;
	    }
    }
    signal(SIGCHLD, childsig);
}

