/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.6.3b (Yorick, Vivriel, Thuzzle, Ilie 1992-97)"
#define SAVEFILEVER 2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"
#include "beam.h"

int select();
int pause();
int gettimeofday();

/* 	function declarations		*/
static void init_env();
static void init_def_file();
static void mainlupe();
#ifndef OMIT_SCHEDULE
static struct timeval *check_schedule(); 
#endif
#ifdef MOTDFILE
static void printmotd();
#endif
static void get_remote_output();
#ifndef OMIT_SESSIONS
static void process_sub_input();
#endif
static void process_input();
static void getuserinput();
static void setsigs();
static void def_error();
static void read_settings();
static void escape_seq();
static void flashparen();
void termsig();
void pipesig();

/*	GLOBALS		*/

char *hostname;
int portnumber;
char *initstr;		/* initial command to run (send) on connect */

FILE *capfile;		/* capture file or NULL */

#ifndef OMIT_SESSIONS
struct consess {
  int active;
  char *id;
} con_sess[MAXFDSCAN];	/* connections indexed by socket descriptors */
#endif

int sockfd = 1;		/* current socket file descriptor */
fd_set fdset;		/* set of descriptors to select() on */

int waiting_for_pid;	/* = pid, when executing #! or editor */
int wait_result;	/* exit status of the process we waited for */

int linemode;		/* line mode flags (LM_* in cancan.h) */

/* booleans */
int mp_active;		/* 1 = we have #identified ourselves */
int new_settings;	/* set this when user defines new alias/action/etc */
int verbose;		/* 1 if command being parsed is user-entered */
int wrap = 1;		/* 1 = word wrap active */
int speedwalk;		/* speedwalk on? */
int verbatim;		/* 1 = don't expand aliases or process semicolons */

/* for line editing */
void (*last_edit_func)();
int edlen;		/* length of current input line */
int pos;		/* cursor position in line */
int flashback;		/* cursor is on excursion, put back there */
char *prefixstr;	/* inserted in the editing buffer each time */
char edbuf[LINELEN]; 	/* line editing buffer */
char inserted_next[LINELEN];

char footnote[LINELEN];	/* displayed at the bottom (just once) */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */

int main(argc, argv)
int argc; char *argv[];
{
    if(argc < 2 || argc > 3) {
	fprintf(stderr, "usage:\t%s host port\nor\t%s definition-file\n",
		argv[0], argv[0]);
	return 1;
    }
    settermcap();
    default_bindings();
    init_env();
#ifndef OMIT_PRINTVER
    printver();
#endif
#ifdef MOTDFILE
    printmotd();
#endif
    if(argc == 2) {	/* one arg, assuming it is definition file name */
	init_def_file(argv[1]);
    } else {	/* two args: assume they are hostname and port number */
	hostname = argv[1];
	portnumber = atoi(argv[2]);
	keypad_bindings();
    }
    default_completions();
    set_terminal();
    setsigs();
    mainlupe();
    return 1;
}

/*
 * initialize cancan from definition file or create one
 */
static void init_def_file(df)
char *df;
{
    char buf[LINELEN], *p;
    FILE *f;
    /*
     * rules: if cancan dir is set it is searched first.
     *        if file doesn't exist, it is created there.
     * If a slash appears in the name, the cancan dir isn't used.
     */
    if(!strchr(df, '/') && (p = getenv("CANCANDIR")) && *p) {
	/* BUG: possible buffer overflow. (snicker) */
	sprintf(buf, "%s/%s", p, df);
	if((f = fopen(buf, "r")) || !(f = fopen(df, "r")))
	    df = buf;
    } else
	f = fopen(df, "r");
    set_env("SAVEFILE", df);
    if(f) {
	read_settings(f);
    } else {
	if(!(f = fopen(df, "a+")))
	    exit(syswarn("Could not open definition file"));
	fclose(f);
	printf("Creating %s\nHost name  :", df);
	fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	if(*skipspace(buf) != '\n') {
	    chop_str(buf, '\n');  hostname = str_dup(buf);
	    printf("Port number:");
	    fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	    portnumber = atoi(buf);
	}
	keypad_bindings();
	save_settings();
    }
}

/*
 * show current version
 */
void printver()
{
    printf(
#if __STDC__
	"cancan ver " VERSION "\nCompiled " __DATE__ " (%s %s)\n",
#else
	"cancan ver %s (%s %s)\n", VERSION,
#endif
#ifdef VT100
	"vt100-only",
#else
	"termcaps",
#endif
#ifndef USESGTTY
	"termio"
#else
	"BSD sgtty"
#endif
    );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
static void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, LINELEN, f))
	    fputs(line, stdout);
    }
}
#endif


/*
 * main loop.
 */
static void mainlupe()
{
    int err;
    fd_set readfds;
    struct timeval timeoutbuf, *timeout;
    
    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
#ifndef OMIT_SESSIONS
    if(hostname && connect_sess((char *)0, initstr, (char *)0, 0) < 0)
	exitcancan("");
#else
    if(!hostname || (sockfd = doconnect(hostname, portnumber)) < 0)
	exitcancan("");
    FD_SET(sockfd, &fdset);	/* add the socket to select() set */
    if(initstr) parse(initstr, 0, (struct param *)0);
#endif
    
    printf("Type #help for help.\n");
    
    /* make sure the input line is always the bottom one */
    check_winsize();
    lastline();
    
    /*		main lupe	*/
    
    for(;;) {
#ifndef OMIT_SCHEDULE
	timeout = check_schedule(&timeoutbuf);
	if(flashback && (!timeout || timeout->tv_sec)) {
#else
	timeout = 0;
	if(flashback) {
#endif
	    timeout = &timeoutbuf;
	    timeout->tv_sec = FLASHDELAY / 1000000;
	    timeout->tv_usec = FLASHDELAY % 1000000;
	}

	if(footnote[0]) display_footnote(1);

	flush_output_buffer();	/* sploooosh */
	fflush(stdout);		/* ditto */

	readfds = fdset;
	err = select(MAXFDSCAN, &readfds, (fd_set *)0, (fd_set *)0, timeout);
	if(err < 0 && errno != EINTR) syserr("select");

	if(flashback) { moveto(flashback);  flashback = 0; }
	if(footnote[0]) display_footnote(0);

	if(err > 0) {
#ifndef OMIT_SESSIONS
	    if(FD_ISSET(0, &readfds)) {
		sockfd = main_sess();
		getuserinput();
		if(!--err) continue;
	    }
	    for(sockfd = 3; sockfd < MAXFDSCAN; sockfd++)
		if(FD_ISSET(sockfd, &readfds)) {
		    get_remote_output();
		    if(!--err) break;
		}
#else
	    if(FD_ISSET(0, &readfds))
		getuserinput();
	    if(FD_ISSET(sockfd, &readfds))
		get_remote_output();
#endif
	}
    }
}

#ifndef OMIT_SCHEDULE
/*
 * process and update delayed commands list.
 */
static struct timeval *check_schedule(to) 
struct timeval *to;
{
    long delta;
    static struct timeval time0;
    if(schedule) {
	delta = -time0.tv_sec;
	gettimeofday(&time0, (struct timezone *)0);
	time0.tv_sec = SCH_HZ * time0.tv_sec +
		time0.tv_usec / (int)(1000000 / SCH_HZ);
	if(delta) delta += time0.tv_sec;
	while(schedule && schedule->delta <= delta)
	    if(schedule->delta >= 0) {
		delta -= schedule->delta;
		schedule->delta = -1;
		if(schedule->next)
		    schedule->next->delta++;
		run_command(schedule->commands);
	    } else
		delete_schedulenode(&schedule);
	if(schedule) {
	    schedule->delta -= delta;
	    to->tv_sec = schedule->delta / SCH_HZ;
	    to->tv_usec = (1000000 / SCH_HZ) * (schedule->delta % SCH_HZ);
	    return to;
	}
    }
    time0.tv_sec = 0;
    return 0;
}
#endif

/*
 * get output from the socket and process/display it.
 */
static void get_remote_output()
{
    static char buffer[BUFSIZE];
    int got = read_socket(buffer, sizeof(buffer) - 1);
    if(!got) return;		/* leave room ^^^ for terminating zero */
#ifdef DEBUGCODE
    /* debug code to see in detail what strange codes come from the server */
    {
	int i;
	for(i = 0; i < got; i++) 
	    fprintf(stderr, isprint(buffer[i]) ? "%c" : "[%d]", buffer[i]);
    }
#endif
  
#ifndef OMIT_SESSIONS
    if(!con_sess[sockfd].active) return;  /* process only active sessions */
#endif
    if(linemode & LM_CHAR) {
	while(write(1, buffer, got) < 0)
	    if(errno != EINTR) { syserr("write"); break; }
    } else {
	/* line-at-a-time mode: process input in a number of ways */
	if(pager_text) pager(0);
	else clearinpline();
	hilite(0);
#ifndef OMIT_SESSIONS
	if(!is_main_sess()) process_sub_input(buffer, got);
	else
#endif
	process_input(buffer, got);
	redraw_input();
    }
}

#ifndef OMIT_SESSIONS
/*
 * process input from remote host (subsidiary session):
 * assert(sizeof(buf) >= size + 1)
 */
static void process_sub_input(buf, size)
char *buf; int size;
{
    char *line, *p;
    int i;

    clear_prompt();
    for(i = 0; i < size; i++) {
	if(buf[i] == '\r') continue;
	for(p = line = buf + i; i < size && buf[i] != '\n'; i++)
	    if(buf[i]) *p++ = buf[i];	/* cut out NUL's */
	*p = '\0';		/* there is room */
	/* FEATURE: no wrapping, no marking in subsidiary session */
	/* BUG: escape codes not removed */
	if(search_action(line)) {
	    printf("%.*s",
		    cols_1 - printf("##%s> ", con_sess[sockfd].id), line);
	    newline();
	}
    }
    redraw_prompt();
}
#endif

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 * assert(sizeof(buf) >= size + 1)
 */
static void process_input(buf, size)
char *buf; int size;
{
    static editsess *esess;
    char *p;
    int i = 0;

    if(data_sent) {
	/*
	 * FEATURE: in case we are given no text to print (only an mp
	 * message, for example), we resurrect the prompt. This happens
	 * to be convenient since mp messages arent followed by a prompt
	 */
	data_sent = 0;
	if(esess) esess->mode = 0;	/* sending cancels the session */
    }
    while(i < size) {
	p = buf + i;
	if(esess || (mp_active && !col0 && !strncmp(p, MPI, strlen(MPI))))
	    i += mp_receive(&esess, p, size - i);
	else {
	    /* process block one line at a time: */
	    if(!col0 && *p == '\r') i++;		/* leading CR */
	    for(p = buf; i < size && buf[i] != '\n'; i++)
		if((*p = buf[i])) p++;			/* cut out NUL's */
	    if(p != buf && *(p - 1) == '\r') p--;	/* trailing CR */
	    *p = '\0';
	    if(capfile) fputs(buf, capfile);
	    if(i++ < size) {
		if(capfile) putc('\n', capfile); 
		if(!search_action(buf)) continue;	/* line #gagged */
		if(buf[0])
		    if(wrap) wrap_out(buf);
		    else mark_and_print(buf);
		promptstr[0] = '\0';
		col0 = 0;  newline();
	    } else {
		/*
		 * FEATURE: action search on line fragments can give results
		 * differing from a search on a complete line. Alternatively,
		 * the search could be postponed until a newline is read.
		 */
		search_action(buf);
		if(!(p = wrap_out(buf))) p = buf;
		else promptstr[0] = '\0';
		if(col0 < cols_1 && strlen(promptstr) + strlen(p) < LINELEN)
		    strcat(promptstr, p);		/* promptocatenate */
		else {
		    col0 = 0;  newline();
		}
		break;
	    }
	}
    } /* end while */
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
static void getuserinput()
{
    unsigned char c;
    static char typed[CAPLEN];	/* chars typed so far (with partial match) */
    static int nchars = 0;	/* number of them */
    keynode *p;
    /*
     * We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line more, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     */
    if(read(0, &c, 1) != 1)
	syserr("read from stdin");
    if(pager_text && pager(c)) return;
    if(linemode & (LM_CHAR | LM_NOECHO)) {
	if(linemode & LM_CHAR) {
	    while(write(sockfd, &c, 1) < 0)	/* blocking i/o */
		if(errno != EINTR) { syserr("write to socket"); break; }
	    if(!(linemode & LM_NOECHO))
		putchar(c);
	} else {
	    /* sending password (line mode, no echo) */
	    edbuf[edlen++] = c;
	    if(edbuf[0] == '\003') intrsig();
	    if((c == '\n' && (newline(), 1)) || edlen >= LINELEN - 2) {
		send_to_host(edbuf, edlen);
		edlen = 0;
	    }
	    edbuf[pos = edlen] = 0;
	}
    } else {	/* normal mode (echo, line buffered) */
	/* NOTE: we should ignore a NUL char */
#ifndef VT100
	if(metakey && (c & 0x80)) {	/* did meta key set the high bit? */
	    typed[nchars++] = '\033';	/* prepend esc */
	    c &= 0x7f;
	}
#endif
	if(!nchars && isprint(c)) {
	    if(edlen >= LINELEN - 1) return;
	    hilite(1);
	    insertchars(&c, 1);
	    if(isrparen(c)) flashparen(c);
	    last_edit_func = 0;
	} else {
	    /* search function key strings for match */
	    typed[nchars++] = c;
	    if((p = keydefs))
		while(strncmp(typed, p->sequence, nchars) && (p = p->next));
	    if(p) {
		if(p->sequence[nchars]) return;		/* partial match */
		p->funct(p->call_data);
		last_edit_func = p->funct;
	    }
	    nchars = 0;
	}
    }
}

/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
static void flashparen(c)
unsigned char c;
{
    int lev = 0, i = pos;
    while(--i >= 0)
	if(isrparen(edbuf[i])) {
	    lev++;
	} else if(islparen(edbuf[i]) && !--lev) {
	    if(LPAREN(c) == edbuf[i]) {
		flashback = pos;
		moveto(i);
	    }
	    break;
	}
}

#ifdef NEWBIE
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
  char *p, *q, *r;
  char buf[LINELEN];
  int n;
  int strncasecmp();
  
  strcpy(buf, edbuf);
  p = buf; r = edbuf;
  while(*p) 
   {
     while(*p && !islabel(*p)) *r++ = *p++;
     q = p;
     while(*p && islabel(*p)) p++;
     if(*p || *q) 
      {
	n = p - q;
	if(strncasecmp(q, "u", n) == 0) 
	 {
	   strcpy(r, "you"); r += 3;
	 }
	else if(strncasecmp(q, "ur", n) == 0) 
	 {
	   strcpy(r, "your"); r += 4;
	 }
	else if(strncasecmp(q, "urs", n) == 0) 
	 {
	   strcpy(r, "yours"); r += 5;
	 }
	else if(strncasecmp(q, "r", n) == 0) 
	 {
	   strcpy(r, "are"); r += 3;
	 }
	else if(strncasecmp(q, "4", n) == 0) 
	 {
	   strcpy(r, "for"); r += 3;
	 }
	else if(strncasecmp(q, "2", n) == 0) 
	 {
	   strcpy(r, "to"); r += 2;
	 }
	else if(strncasecmp(q, "b", n) == 0) 
	 {
	   strcpy(r, "be"); r += 2;
	 } 
	else if(strncasecmp(q, "b4", n) == 0) 
	 {
	   strcpy(r, "before"); r += 6;
	 }
	else 
	 {
	   strcpy(r, q); r += n;
	 }
      }
   }
  *r = '\0';
  edlen = r - edbuf;
}
#endif

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan(message)
char *message;
{
    reset_terminal();
    printf("%s#cancan stopped dancing.\n", message);
    exit(0);
}
 
#ifndef NOSHELL
/*
 * execute a shell command and return the result
 */
int execute(arg)
char *arg;
{
    flush_output_buffer();
    fflush(stdout);
    if(!(waiting_for_pid = fork())) {
	static char *args[] = { "/bin/sh", "-c", 0, 0 };
	args[2] = arg;
	execve(args[0], args, environ);
	_exit(1);
    } else if(waiting_for_pid < 0) {
	printf("#fork failed, sorry\n");
    } else {
	while(waiting_for_pid) pause();
    }
    return wait_result;
}
#endif

/*
 * set up our signal handlers
 */
static void setsigs()
{
    signal(SIGCHLD, childsig);
    signal(SIGTERM, termsig);
    signal(SIGPIPE, pipesig);
    signal(SIGWINCH, winchsig);
}

void intrsig()
{
    if(sockfd > 1) close(sockfd);
    exitcancan("\n#interrupted.\n");
}

void pipesig()
{
    exitcancan("\n#broken pipe.\n");
}

void termsig()
{
    exitcancan("\n#termination signal.\n");
}

/*
 * suspend ourselves
 */
void suspend_cancan()
{
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	reset_terminal();  fflush(stdout);
	signal(SIGTERM, SIG_IGN);
	signal(SIGTSTP, SIG_DFL);
	kill(getpid(), SIGTSTP);
	set_terminal();
	signal(SIGTERM, termsig);
	check_winsize();	/* in case size changed meanwhile */
	lastline();		/* so we know where the cursor is */
	redraw_prompt();
	redraw_input();
    } else
	strcpy(footnote, "#your shell probably lacks job control.");
}

/*
 * print an error message
 */
int syswarn(msg)
char *msg;
{
    if(errno > 0 && errno < sys_nerr)
	printf("%s (%s)\n", msg, sys_errlist[errno]);
    else
	printf("%s\n", msg);
    return -1;
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    reset_terminal();
    fprintf(stderr, "\n#fatal system call error: %s (", msg);
    if(errno > 0 && errno < sys_nerr)
	fprintf(stderr, "%s)\n", sys_errlist[errno]);
    else
	fprintf(stderr, "%d)\n", errno);
    exit(1);
}

static void def_error(ln)
int ln;
{
    static int n_err;
    printf("bad definition file format, line %d\n", ln);
    if(++n_err >= 10) exit(1);
}

/*
 * read definitions from file
 */
static void read_settings(f)
FILE *f;
{
    char buf[LINELEN];
    char *p, *cmd;
    int ln;

    for(ln = 1; fgets(buf, sizeof(buf), f); ln++) {
	if(*buf == '\n') continue;
	p = chop_str(cmd = buf, ' ');
	if(p && *cmd == '#') {
	    chop_str(p, '\n');	/* strip lf */
	    if(!strcmp(++cmd, "action"))
		parse_action(p);
	    else if(!strcmp(cmd, "alias"))
		parse_alias(p);
	    else if(!strcmp(cmd, "bind"))
		parse_bind(p);
	    else if(!strcmp(cmd, "mark"))
		parse_marker(p);
	    else if(!strcmp(cmd, "init"))
		str_assign(&initstr, p);
#if 0
	    else if(!strcmp(cmd, "hilite"))
		attr_string(parse_attributes(p), edattrbeg, edattrend);
#endif
	    else if(!strcmp(cmd, "savefile")) {
		if(atoi(p) > SAVEFILEVER)
		    printf("Warning: this cancan version is too old!\n");
	    } else if(!strcmp(cmd, "host")) {
		char *rest = chop_str(p, ' ');
		if(rest && (portnumber = atoi(rest)))
		    str_assign(&hostname, p);
		else
		    def_error(ln);
	    } else
		def_error(ln);
	} else
	    def_error(ln);
    }
    fclose(f);
}

/*
 * make an escape sequence printable (for use in definition-files)
 */
static void escape_seq(seq, buf)
char *seq, *buf;
{
    char *p = buf;
    /*
     * rules: escape =\^ with \. control chars are written as ^X.
     */
    for(;*seq; seq++) {
	if(*seq < ' ') {
	    *p++ = '^';
	    *p++ = *seq | '@';
	    continue;
	}
	if(*seq == '=' || *seq == '\\' || *seq == '^')
	    *p++ = '\\';
	*p++ = *seq;
    }
    *p = '\0';
}

/*
 * save settings in definition file
 */
void save_settings()
{
    char buf[LINELEN], *cp;
    FILE *f;
    aliasnode *al;
    actionnode *ac;
    marknode *m;
    keynode *k;
  
    if(!(cp = getenv("SAVEFILE"))) return;
    if(!(f = fopen(cp, "w"))) {
	printf("#error: can't write to definition file %s\n", cp);
	return;
    }
    fprintf(f, "#savefile %d\n", SAVEFILEVER);
    if(hostname)
	fprintf(f, "#host %s %d\n", hostname, portnumber);
    if(initstr)
	fprintf(f, "#init %s\n", initstr);
#if 0
    if(edattr != NOATTRCODE)
	fprintf(f, "#hilite %s\n", attr_name(edattr));
#endif
    for(al = aliases; al; al = al->next)
	fprintf(f, "#alias %s=%s\n", al->name, al->subst);
    for(ac = actions; ac; ac = ac->next) {
	escape_specials(ac->pattern, buf, "\\\"");
	fprintf(f, "#action %c%s \"%s\"=%s\n", (ac->active) ? '+' : '-',
		ac->label, buf, ac->commands);
    }
    for(m = markers; m; m = m->next) {
	escape_specials(m->pattern, buf, "\\=");
	fprintf(f, "#mark %s=%s\n", buf, attr_name(m->attrcode));
    }
    /* FEATURE: deleted default keys should be saved too, as "#bind <key>=" */
    for(k = keydefs; k; k = k->next)
	if(k->call_data) {
	    escape_seq(k->sequence, buf);
	    fprintf(f, "#bind %s %s=%s\n", k->name, buf, k->call_data);
	}
    fclose(f);
}

/*
 * if alias recording is active, record given line in that alias
 */
void record_line(line)
char *line;
{
    static int i;
    aliasnode *np;
    if(!(np = *lookup_alias(recording_name))) {
	i = 0;
	add_aliasnode(recording_name, line);
    } else {
	char new_line[LINELEN];
	int len = sprintf(new_line, "%s", np->subst);
	/*
	 * FEATURE: if both alias name and the line to be recorded
	 * are very long, total alias lenght may grow too large.
	 */
	int room = LINELEN - 10 - strlen(line);
	if(len >= room - LINELEN / 2) {	/* leave ample room for editing */
	    do {
		len = sprintf(new_line, "%s%d", recording_name, ++i);
	    } while(*lookup_alias(new_line));
	    if(len >= room) return;
	    add_aliasnode(new_line, np->subst);
	}
	sprintf(new_line + len, ";%s", line);
	str_assign(&np->subst, new_line);
    }
}

char *str_dup(s)
char *s;
{
    char *k = (char *)malloc(strlen(s) + 1);
    if(!k) syserr("malloc");
    return strcpy(k, s);
}

void str_assign(str, s)
char **str, *s;
{
    if(*str) free(*str);
    *str = s ? str_dup(s) : 0;
}

#ifndef OMIT_SESSIONS
/*
 * Below are multiple-session support functions:
 *
 * note on con_sess[MAXFDSCAN]:
 *	con_sess[0].active holds total no. of connections
 *	con_sess[1].active holds main socket descriptor
 */

/*
 * return TRUE if current session being processed is main connection
 */
int is_main_sess()
{
    return (sockfd == con_sess[1].active);
}

/*
 * return main session sockfd
 */
int main_sess()
{
    return con_sess[1].active;
}

/*
 * return session's index or socket descriptor given id,
 * or 0 if invalid id is given
 */
int find_sess(id, notify)
char *id;
int notify;
{
    int i, n;
    for(i = 3, n = con_sess[0].active; n; i++)
	if(con_sess[i].id) {
	    n--;
	    if(!strcmp(con_sess[i].id, id))
		return i;
	}
    if(notify)
	printf("#no such connection\n");
    return 0;
}

/*
 * connect another session
 */
int connect_sess(id, init, host, port)
char *id;	/* session id - NULL if main session */
char *init;	/* optional init string */
char *host;
int port; 
{
    int fd;
    if((!host && !(host = hostname)) || (!port && !(port = portnumber))) {
	printf("#please specify both hostname and port number.\n");
	return -1;
    }
    if(init && !*init) init = 0;
    /* dial the number by moving the right index in small circles */
    if((fd = doconnect(host, port)) < 0)
	return fd;
    FD_SET(fd, &fdset);		/* add socket to select() set */
    con_sess[0].active++;
    if(!con_sess[1].active)	/* save main socket fd */
	con_sess[1].active = sockfd = fd;
    if(!id) {			/* main session */
	con_sess[fd].active = 1;
	con_sess[fd].id = "main";
    } else {
	con_sess[fd].active = !init;
	con_sess[fd].id = str_dup(id);
    }
    if(init) parse(init, 0, (struct param *)0);
    return fd;
}

/*
 * disconnect session
 */
void disconnect_sess(id)
char *id;
{
    int sfd;

    if(id) {				/* #disconnect cmd */
	if(!(sfd = find_sess(id, 1))) return;
	close(sfd);
    } else
	sfd = sockfd;			/* connection closed by remote host */

    clear_prompt();
    printf("#connection \"%s\" closed.\r\n", con_sess[sfd].id);
    if(sfd == con_sess[1].active)	/* main connection closed */
	exitcancan("");
    redraw_prompt();

    FD_CLR(sfd, &fdset);
    con_sess[0].active--;
    con_sess[sfd].active = 0;
    free(con_sess[sfd].id);
    con_sess[sfd].id = 0;
}

/*
 * toggle output display from another session
 */
void snoop_sess(id)
char *id;
{
    int sfd = find_sess(id, 1);
    if(sfd)
	con_sess[sfd].active ^= 1;
}

/*
 * show open connections
 */
void show_connections()
{
    int n = con_sess[0].active;
    if(!n)
	printf("#No connections have been made\n");
    else {
	int i = 2;
	sprintf(inserted_next, "#The following connection%s open:",
		(n == 1) ? " is" : "s are");
        do {
	    paged_print(inserted_next);
	    while(!con_sess[++i].id);
	    sprintf(inserted_next, "##%-8s%s", con_sess[i].id,
		    !con_sess[i].active ? "" :
		    i == con_sess[1].active ?  " (active)" : " (snooped)");
	} while(--n);
    }
}
#endif

/*
 * initialize our own environment
 */
static void init_env()
{
    char **oldenv = environ;
    int i = 0;
    while(oldenv[i++]);
    environ = (char **)malloc(i * sizeof(char *));
    if(!environ) exit(syswarn("Could not allocate memory"));
    for(environ[--i] = 0; i; --i, environ[i] = str_dup(oldenv[i]));
}

/*
 * set or unset (when value == NULL) an environment variable
 */
void set_env(variable, value)
char *variable, *value;
{
    int i, len = strlen(variable);
    char *k;
    for(i = 0; (k = environ[i]); i++)
	if(!strncmp(k, variable, len) && (!k[len] || k[len] == '='))
	    break;
    if(value) {
	static int envsize;
	char buf[LINELEN];
	sprintf(buf, "%s=%s", variable, value);
	str_assign(environ + i, buf);
	if(k) return;
	if(++i >= envsize) {
	    envsize = (i | 63) + 1;
	    environ = (char **)realloc(environ, envsize * sizeof(char *));
	    if(!environ) syserr("realloc");
	}
	environ[i] = 0;
    } else if(k) {
	free(k);
	do i++; while((environ[i - 1] = environ[i]));
    }
}

