/*
 * cmd.c
 * functions for cancan's built-in #commands
 *
 * (created: Finn Arne Gangstad, Dec 25th, 1993)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef AIX
#include <sys/select.h>
#endif

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"
#include "beam.h"

FILE *popen();
int pclose();
int select();

static void show_aliases();
static void show_actions();
static void show_keybinds();
static char *unescape_seq();
static char *bindtext_by_funct();
static void keybind_modify();
static int key_collision();
static char *read_seq();
static int get_one_char();
static void print_seq();
static void show_marks();
static int parse_attributes();
static int attr_token();
#ifndef OMIT_SCHEDULE
static void show_schedule();
static void parse_schedule();
static long parse_delay();
#endif
static void show_variables();
static int toggle();

static void cmd_action(), cmd_alias(), cmd_bind(), cmd_cancel();
static void cmd_capture(), cmd_echo(), cmd_edit(), cmd_for();
static void cmd_help(), cmd_hilite(), cmd_identify(), cmd_init();
static void cmd_map(), cmd_mark(), cmd_prefix(), cmd_quote();
static void cmd_read(), cmd_record(), cmd_retrace(), cmd_set();
static void cmd_speedwalk(), cmd_ver(), cmd_wrap();
#ifndef OMIT_SESSIONS
static void cmd_connect(), cmd_snoop(), cmd_zap();
#endif
#ifndef OMIT_SCHEDULE
static void cmd_schedule();
#endif
#ifndef NOSHELL
static void cmd_shell();
#endif
#ifdef TELNETBUG
static void cmd_stdcolour();
#endif

const cmdstruct commands[] = 
{
   { "help", "[topic]\t\t\thelp on commands, or this text", cmd_help },
   { "17", "command\t\t\trepeat 'command' 17 times", 0 },
#ifndef NOSHELL
   { "!", "shell-command\t\texecute a shell command using /bin/sh", cmd_shell },
   { "?", "shell-command;...[;...]\tif .. then .. else", 0 },
#endif
   { "action", "[name][\"pattern\"][=cmd]\tlist/delete/define actions",
	cmd_action },
   { "alias", "[name[=text]]\t\tlist/delete/define aliases", cmd_alias },
   { "bind", "[name[=command]]\t\tlist/delete/define key bindings", cmd_bind },
   { "cancel", "[number]\t\tcancel editing session", cmd_cancel },
   { "capture", "[filename]\t\tbegin/end of capture to file", cmd_capture },
#ifndef OMIT_SESSIONS
   { "connect", "[id[ init][=address]]\tconnect a new session",
	cmd_connect },
#endif
   { "echo", "[-n ][text]\t\tdisplay text on screen", cmd_echo },
   { "edit", "\t\t\t\tlist editing sessions", cmd_edit },
   { "for", "a b [c] cmd|var cmd\titerate a command", cmd_for },
   { "hilite", "[attr]\t\t\thighlight your input line", cmd_hilite },
   { "identify", "[startact endact]\tsend MUME client identification",
	cmd_identify },
   { "init", "[-][[+]string]\t\tsend string to host on connect", cmd_init },
   { "map", "[[-]number]\t\tshow/clear (auto)map", cmd_map },
   { "mark", "[string[=attr]]\t\tlist/define markers", cmd_mark },
   { "prefix", "string\t\t\tprefix all lines with string", cmd_prefix },
   { "quote", "[on|off]\t\t\ttoggle verbatim-flag", cmd_quote },
   { "read", "file\t\t\tread lines from file/pipe and send", cmd_read },
   { "record", "[name]\t\t\trecord typed lines into alias", cmd_record },
   { "retrace", "[[-]number]\t\tretrace the last number steps", cmd_retrace },
#ifndef OMIT_SCHEDULE
   { "schedule", "name[ delay][=value]\tlist/delete/define delayed commands",
	cmd_schedule },
#endif
   { "set", "name[=value]\t\tset variable value", cmd_set },
#ifndef OMIT_SESSIONS
   { "snoop", "id\t\t\ttoggle output display for session", cmd_snoop },
#endif
   { "speedwalk", "[on|off]\t\ttoggle speedwalk mode", cmd_speedwalk },
#ifdef TELNETBUG
   { "stdcolour", "attr\t\t\tset standard background colour", cmd_stdcolour },
#endif
   { "ver", "\t\t\t\tshow cancan version", cmd_ver },
   { "wrap", "[on|off]\t\t\tturn word wrapping on/off", cmd_wrap },
#ifndef OMIT_SESSIONS
   { "zap", "id\t\t\t\tdisconnect a session", cmd_zap },
#endif
   { 0, 0, 0 }
};

/* anyone knows if ANSI 6429 talks about more than 8 colours? */
static const char colournames[][8] = {
    "black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"
};

char *recording_name;		/* name of recording alias or NULL */

static void cmd_action(arg)
char *arg;
{
    if(!*arg)
	show_actions();
    else
	parse_action(arg);
}

/*
 * show defined actions
 */
static void show_actions()
{
    actionnode *p;
  
    sprintf(inserted_next, "#%s action%s defined%c",
	    actions ? "The following" : "No",
	    (actions && !actions->next) ? " is" : "s are",
	    actions ? ':' : '.');
    for(p = actions; paged_print(inserted_next), p; p = p->next)
	sprintf(inserted_next, "#action %c%s \"%s\"=%s",
		p->active ? '+' : '-', p->label, p->pattern, p->commands);
}

/*
 * parse the #action command
 */
void parse_action(str)
char *str;
{
    actionnode **np = 0;
    char *label = 0, *pattern = 0, *rest = 0;
    char *s = str, sign = 0;
    if(*s != '"' && (
	    label = *s != '-' && *s != '+' ? s : (sign = *s, s + 1),
	    s = chop_arg(label, ' '),
	    rest = chop_arg(label, '='))) {
	if(s) *(s - 1) = ' ';
    } else if(s) {
	if(*s == '"') {
	    if(!(s = chop_arg(pattern = s + 1, '"'))) {
		printf("#action: pattern not \" terminated\n");
		return;
	    }
	    unescape(pattern);
	}
	if(*(s = skipspace(s)))
	    if(*s != '=') {
		printf("#action: missing a '=' after label or pattern\n");
		return;
	    } else
		rest = s + 1;
    }
    /* at this point, either label or pattern or both are set */
    np = label ? lookup_action(label) : lookup_action_pattern(pattern);
    /*
     * FEATURE: we do not check if two actions have the same pattern.
     * Should we warn then? Deactivate the second one? hmm.
     */
    if(*np) {
	if(!rest) {
	    if(sign) {
		(*np)->active = sign == '+';
		if(verbose)
		    printf("#action %s \"%s\" is now o%s\n", (*np)->label,
			    (*np)->pattern, (sign == '+') ? "n" : "ff");
	    } else {
		rest = inserted_next;
		rest += sprintf(rest, "#action %s%s \"",
			(*np)->active ? "" : "-", (*np)->label);
		escape_specials((*np)->pattern, rest, "\\\";");
		sprintf(rest + strlen(rest), "\"=%s", (*np)->commands);
	    }
	} else if(!*rest) {
	    if(verbose)
		printf("#action %s \"%s\" deleted\n",
			(*np)->label, (*np)->pattern);
	    delete_actionnode(np);
	    new_settings = 1;
	} else {
	    if(sign && (sign = (*np)->active != (sign == '+')))
		(*np)->active ^= 1;
	    if(pattern && strcmp((*np)->pattern, pattern)) {
		str_assign(&(*np)->pattern, pattern);
		sign = 1;
	    }
	    if(strcmp((*np)->commands, rest)) {
		str_assign(&(*np)->commands, rest);
		sign = 1;
	    }
	    if(sign) {
		if(verbose) printf("#changed action: %s%s \"%s\"=%s\n",
			(*np)->active ? "" : "-", (*np)->label,
			(*np)->pattern, (*np)->commands);
		new_settings = 1;
	    }
	}	    	    
    } else if(!pattern) {
	printf("#no action labelled '%s'\n", label);
    } else if(!rest || !*rest) {
	printf("#no action with pattern \"%s\"\n", pattern);
    } else {
	static int last = 0;
	static char lbuf[16];
	if(!label) {
	    do sprintf(lbuf, "%d", ++last); while(*lookup_action(lbuf));
	    label = lbuf;
	}
	np = add_actionnode(pattern, rest, label, sign != '-');
	if(verbose)
	    printf("#new action: %s \"%s\"=%s\n", label, pattern, rest);
	new_settings = 1;
    }
}

static void cmd_alias(arg)
char *arg;
{
    if(!*arg)
	show_aliases();
    else
	parse_alias(arg);
}

/*
 * show defined aliases
 */
static void show_aliases()
{
    aliasnode *p;
  
    sprintf(inserted_next, "#%s alias%s defined%c",
	    aliases ? "The following" : "No",
	    (aliases && !aliases->next) ? " is" : "es are",
	    aliases ? ':' : '.');
    for(p = aliases; paged_print(inserted_next), p; p = p->next)
	sprintf(inserted_next, "#alias %s=%s", p->name, p->subst);
}

/*
 * parse the #alias command
 */
void parse_alias(str)
char *str;
{
    aliasnode **np;
    char *rest = chop_arg(str, '=');
    if(chop_arg(str, ' ')) {
	printf("#missing a '=' after alias label.\n");
	return;
    }
    np = lookup_alias(str);
    if(*np) {
	if(!rest)
	    sprintf(inserted_next, "#alias %s=%s", str, (*np)->subst);
	else if(!*rest) {
	    if(verbose)
		printf("#deleting alias: %s=%s\n", str, (*np)->subst);
	    delete_aliasnode(np);
	    new_settings = 1;
	} else if(strcmp(rest, (*np)->subst)) {
	    str_assign(&(*np)->subst, rest);
	    if(verbose)
		printf("#changed alias: %s=%s\n", str, rest);
	    new_settings = 1;
	}
    } else if(rest && *rest) {
	if(strcmp(str, "#alias")) {
	    add_aliasnode(str, rest);
	    if(verbose)
		printf("#new alias: %s=%s\n", str, rest);
	    new_settings = 1;
	} else
	    printf("#won't alias \"#alias\"\n");
    } else
	printf("#unknown alias: %s\n", str);
}

static void cmd_bind(arg)
char *arg;
{
    if(!*arg)
	show_keybinds();
    else
	parse_bind(arg);
}

/*
 * list keyboard bindings
 */
static void show_keybinds()
{
    keynode *p;
    int count = 0;
    for(p = keydefs; p; p = p->next)
	if(p->call_data) {
	    if(!count++)
		paged_print("#user-defined keys:");
	    sprintf(inserted_next, "#bind %s=%s", p->name, p->call_data);
	    paged_print(inserted_next);
	}
    if(!count)
	printf("#No user-defined key bindings.\n");
}

/*
 * parse the #bind command 
 */
void parse_bind(str)
char *str;
{
    keynode **np;
    char *text;
    char *rest = chop_arg(str, '=');
    char *kseq = chop_arg(str, ' ');

    if(kseq && !*(kseq = unescape_seq(skipspace(kseq)))) kseq = 0;
    np = lookup_key(str);
    if(*np) {
	if(!(text = (*np)->call_data) &&
		!(text = bindtext_by_funct((*np)->funct))) return;
	if(kseq && strcmp(kseq, (*np)->sequence))
	    printf("#key '%s' already bound; will not rebind.\n", str);
	else if(!rest)
	    sprintf(inserted_next, "#bind %s=%s", str, text);
	else if(!*rest) {
	    if(verbose)
		printf("#key binding deleted: %s=%s\n", str, text);
	    delete_keynode(np);
	    new_settings = 1;
	} else if(strcmp(rest, text)) {
	    (*np)->funct = run_command;
	    keybind_modify(*np, rest);
	    if(verbose)
		printf("#redefined key: %s=%s\n", str, rest);
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#no such key: '%s'\n", str);
    } else if(kseq ? !key_collision(kseq, 1) :
	    !key_collision(kseq = read_seq(), 0)) {
	np = add_keynode(str, kseq, run_command, (char *)0);
	keybind_modify(*np, rest);
	if(verbose)
	    printf("#new key binding: %s=%s\n", str, rest);
	new_settings = 1;
    }
}

/*
 * unescape, but also recognize ^c syntax.
 */
static char *unescape_seq(str)
char *str;
{
    char *s = str, *d = str;
    do switch(*s) {
	/* NOTE: we should skip NUL's here (i.e. ^@) */
	case '^': *(s + 1) = CONTROL(*(s + 1));
	/* FALLTHROUGH */
	case '\\': s++;
    } while((*d++ = *s++));
    return str;
}

/*
 * given a pointer to edit function, return the name of the function:
 * e.g. if(f == prev_char) return "&prev-char"
 */
static char *bindtext_by_funct(f)
void (*f)();
{
    const struct defbind_node *p = defbinds;
    while(p->funct && p->funct != f) p++;
    if(!p->name)
	printf("#key '%s' is predefined and cannot be modified.\n", p->label);
    return p->name;
}

/*
 * assign keybind's call_data and change funct if necessary.
 * if keybind is found to be a default one, null call_data.
 */
static void keybind_modify(np, text)
keynode *np;
char *text;
{
    const struct defbind_node *p;
    if(*text == '&')
	for(p = defbinds; p->label[0]; p++)
	    if(p->name && !strcmp(p->name, text)) {
		np->funct = p->funct;
		if(p->seq && !strcmp(np->sequence, p->seq) &&
			!strcmp(np->name, p->label)) {
		    text = 0;  break;
		}
	    }
    str_assign(&np->call_data, text);
}

/*
 * determine if a keysequence collides with that of another key.
 */
static int key_collision(keyseq, shadow)
char *keyseq;
int shadow;
{
    keynode **np;
    if(!*keyseq || isprint(*keyseq)) {
	printf("#that is not a redefineable key.\n");
	return 1;
    }
    for(np = &keydefs; *np; np = &(*np)->next) {
	register char *a = (*np)->sequence;
	register char *b = keyseq;
	do if(!*a || !*b) {
	    if(*a || *b) {
		printf("#colliding keysequence: key '%s'\n", (*np)->name);
	    } else if((a = (*np)->call_data) ||
		    ((a = bindtext_by_funct((*np)->funct)) && !shadow)) {
		printf("#key already bound as: %s=%s\n", (*np)->name, a);
	    } else if(a) {
		delete_keynode(np);  return 0;
	    }
	    return 1;
	} while(*a++ == *b++);
    }
    return 0;
}

/*
 * prompt the user to press a key, read a single escape sequence from the
 * keyboard, print it. return static pointer to sequence.
 */
static char *read_seq()
{
    static char kbuf[CAPLEN];
    int i = 0, timeout = 0;
    printf("#please press the key you want to define:");  fflush(stdout);
    while(i < CAPLEN - 1 && get_one_char(kbuf + i, timeout))
	if(i++ || kbuf[0] != '\033') timeout = 1;
    /* NOTE: we should remove NUL's from the sequence */
    kbuf[i] = 0;
    print_seq(kbuf);  newline();
    return kbuf;
}

/*
 * read a single character from stdin, with timeout in microseconds.
 * timeout == 0 means wait indefinitely (no timeout).
 * return 1 if a char was read, or 0 if not.
 */
static int get_one_char(cp, timeout)
char *cp;
int timeout;
{
    int err;
    fd_set fds;
    struct timeval timeoutbuf;

    FD_ZERO(&fds);	/* this is probably redundant */
    FD_SET(0, &fds);
    timeoutbuf.tv_sec = KBD_TIMEOUT / 1000000;
    timeoutbuf.tv_usec = KBD_TIMEOUT % 1000000;
    while((err = select(1, &fds, (fd_set *)0, (fd_set *)0,
	    timeout ? &timeoutbuf : (struct timeval *)0)) < 0)
	/* NOTE: timeoutbuf contents here could be modified, could be not! */
	if(errno != EINTR) { syserr("select");  break; }
    if(err && read(0, cp, 1) != 1)
	syserr("read from stdin");
    return err;
}

/*
 * print an escape sequence human-readably
 */
static void print_seq(seq)
char *seq;
{
    while(*seq) {
	if(*seq == '\033')
	    printf(" esc");
	else if(*seq == ' ')
	    printf(" space");
	else if(*seq < ' ')
	    printf(" ^%c", *seq | '@');
	else
	    printf(" %c", *seq);
	seq++;
    }
}

static void cmd_cancel(arg)
char *arg;
{
    editsess *sp;
    int num;
    if(!edit_sess) {
	printf("#no active sessions to cancel.\n");
    } else if(*arg) {
	if((num = atoi(arg)))
	    for(sp = edit_sess; sp; sp = sp->next)
		if(num == sp->num) {
		    cancel_edit(sp);
		    return;
		}
	printf("#no such session - use the session number\n");
    } else if(edit_sess->next) {
	printf("#several sessions active, use #cancel <number>\n");
    } else
	cancel_edit(edit_sess);
}

static void cmd_capture(arg)
char *arg;
{
    if(capfile) {
	if(*arg)
	    printf("#capture already active.\n");
	else {
	    fclose(capfile);  capfile = 0;
	    printf("#end of capture to file.\n");
	}
    } else {
	char *mode = "w";
	if(*arg == '>') {
	    arg = skipspace(arg + 1);
	    mode = "a";
	}
	if(!*arg)
	    printf("#capture to what file?\n");
	else if(!(capfile = fopen(arg, mode)))
	    printf("#error writing file %s\n", arg);
	else
	    printf("#capture to %s begins.\n", arg);
    }
}

static void cmd_echo(arg)
char *arg;
{
    char *pat = "%s\n";
    if(arg[0] == '-' && arg[1] == 'n') {		/* #echo -n ... */
	if(!arg[2]) return;
	if(arg[2] == ' ') {
	    pat = "%s";
	    arg += 3;
	}
    }
    printf(pat, unescape_seq(arg));
}

/*ARGSUSED*/
static void cmd_edit(arg)
char *arg;
{
    editsess *s;
    if(!edit_sess)
	printf("#no active editors.\n");
    else for(s = edit_sess; s; s = s->next)
	if(s->state) {
	    sprintf(inserted_next, "#%s session %d \"%s\"",
		    s->mode == 'E' ? "editing" :
		    s->mode == 'V' ? "viewing" : "cancelled",
		    s->num, s->desc);
	    paged_print(inserted_next);
	}
}

static void cmd_for(arg)
char *arg;
{
    /* FEATURE: #for command has no time (or sent data) limit... */
    if(!isdigit(*arg)) {
	char *var, *cmd = chop_arg(arg, ' ');
	int v, last = -1;
	if(cmd) {
	    while((var = getenv(arg)) && (v = atoi(var)) &&
		    (v < 0 ? v = -v : v) < (unsigned)last) {
		last = v;
		parse(cmd, 0, (struct param *)0);
	    }
	    return;
	}
    } else {
	static char buf[32], *s;
	int i, to, step;
	if((i = strtol(arg, &s, 10), *s == ' ') &&
		(to = strtol(s + 1, &s, 10), !*s || *s == ' ')) {
	    if(!*s || !(step = strtol(++s, &arg, 10)) || *arg != ' ')
		step = 1;
	    else
		s = arg + 1;
	    while((step > 0) ? (i <= to) : (i >= to)) {
		sprintf(buf, "%d", i);
		set_env("I", buf);
		parse(s, 0, (struct param *)0);
		i += step;
	    }
	    set_env("I", (char *)0);
	    return;
	}
    }
    printf("#for: bad syntax\n");
}

static void cmd_hilite(arg)
char *arg;
{
    int attr = parse_attributes(arg);
    if(attr >= 0 && verbose) {
	/*
	 * NOTE: #hilite with !verbose (e.g. an action-bound command)
	 * could have some function as well.
	 */
	attr_string(attr, edattrbeg, edattrend);
	edattr = attr - NOATTRCODE;
	printf("#input highlighting is now %so%s%s.\n", 
		edattrbeg, edattr ? "n" : "ff", edattrend);
    }
}

/*ARGSUSED*/
static void cmd_help(arg)
char *arg;
{
    int i;
#ifdef HELPFILE
    if(*arg) {
	FILE *f = fopen(HELPFILE, "r");
	if(!f)
	    printf("#can't open help file %s\n", HELPFILE);
	else {
	    if(*arg == '#') arg++;
	    while(fgets(inserted_next, LINELEN, f)) {
		chop_str(inserted_next, '\n');
		if(inserted_next[0] == '@' && (!inserted_next[1] ||
			!strcmp(arg, inserted_next + 1))) {
		    while(fgets(inserted_next, LINELEN, f) &&
			    inserted_next[0] != '@') {
			chop_str(inserted_next, '\n');
			paged_print(inserted_next);
		    }
		    return;
		}
	    }
	    printf("#no entry for '%s' in the help file.\n", arg);
	}
    } else
#endif
    {
	paged_print("# commands available:");
	for(i = 0; commands[i].name; i++) {
	    sprintf(inserted_next, "#%s %s",
		    commands[i].name, commands[i].help);
	    paged_print(inserted_next);
	}
    }
}

static void cmd_identify(arg)
char *arg;
{
    char *p;
    if(*arg) p = chop_arg(arg, ' ');
    else arg = p = 0;
    str_assign(&edit_start, arg);
    str_assign(&edit_end, p);
    /* NOTE: we ought to be able to reidentify */
    if(!mp_active) {
	mp_active = 1;
	send_to_host(MCIDENT, strlen(MCIDENT));
	data_sent--;		/* MUME doesn't prompt ident string */
    }
}

static void cmd_init(arg)
char *arg;
{
    if(!*arg) {
	if(!initstr)
	    printf("#init not defined.\n");
	else
	    sprintf(inserted_next, "#init %s", initstr);
    } else if(arg[1] || arg[0] != '-') {
	if(!initstr || strcmp(arg, initstr)) {
	    str_assign(&initstr, arg);
	    printf("#init changed.\n");
	    new_settings = 1;
	}
    } else if(initstr) {
	free(initstr);  initstr = 0;
	printf("#init cleared.\n");
	new_settings = 1;
    }
}

static void cmd_map(arg)
char *arg;
{
    if(arg[0] == '+' && arg[1] && strchr("nsewud", arg[1]))
	add_dir_to_map(arg[1]);
    else if(*arg && !*skipnum(arg))
	remove_steps(atoi(arg)); /* remove some steps without walking */
    else
	show_automap();
}

static void cmd_mark(arg)
char *arg;
{
    if(!*arg)
	show_marks();
    else
	parse_marker(arg);
}

/*
 * show defined marks
 */
static void show_marks()
{
    marknode *p;
    sprintf(inserted_next, "#%s marker%s defined%c",
	    markers ? "The following" : "No",
	    (markers && !markers->next) ? " is" : "s are",
	    markers ? ':' : '.');
    for(p = markers; paged_print(inserted_next), p; p = p->next)
	sprintf(inserted_next, "#mark %s=%s",
		p->pattern, attr_name(p->attrcode));
}

/*
 * parse arguments to the #mark command
 */
void parse_marker(str)
char *str;
{
    int attrcode;
    marknode **np;
    char *rest = chop_arg(str, '=');
    unescape(str);
    np = lookup_marker(str);
    if(*np) {
	if(!rest) {
	    strcpy(inserted_next, "#mark ");
	    rest = inserted_next + 6;
	    if(*str == ' ') *rest++ = '\\';
	    escape_specials(str, rest, "\\=;");
	    sprintf(rest + strlen(rest), "=%s", attr_name((*np)->attrcode));
	} else if(!*rest) {
	    if(verbose)
		printf("#mark deleted: %s=%s\n",
			(*np)->pattern, attr_name((*np)->attrcode));
	    delete_marknode(np);
	    new_settings = 1;
	} else if((attrcode = parse_attributes(rest)) >= 0 &&
		(*np)->attrcode != attrcode) {
	    (*np)->attrcode = attrcode;
	    if(verbose)
		printf("#changed mark: %s=%s\n",
			str, attr_name((*np)->attrcode));
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#unknown marker: '%s'\n", str);
    } else if((attrcode = parse_attributes(rest)) >= 0) {
	np = add_marknode(str, attrcode);
	if(verbose)
	    printf("#new mark: %s=%s\n", str, attr_name((*np)->attrcode));
	new_settings = 1;
    }
}

/*
 * return a static pointer to name of given attribute code
 */
const char *attr_name(attrcode)
int attrcode;
{
    static char name[32];	/* should be enough */
    unsigned fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);

    if(bg == NOATTRIB) {
	if(fg < COLOURS) return colournames[fg];
	switch(fg) {
	    case BOLD: return "bold";
	    case UNDERLINE: return "underline";
	    case INVERSE: return "inverse";
	}
    }
    if(fg < COLOURS)
	sprintf(name, "%s on %s", colournames[fg], colournames[bg]);
    else
	sprintf(name, "on %s", colournames[bg]);
    return name;
}

/*
 * parse attribute description in line.
 * Return attribute if successful, -1 otherwise.
 */
static int parse_attributes(line)
char *line;
{
    int t, bg;
    for(t = bg = NOATTRIB; ; ) {
	if(*(line = skipspace(line))) {
	    if((t = attr_token(&line)) < 0) break;
	    if(t <= COLOURS && (t == NOATTRIB || line)) {
		if(t != NOATTRIB && attr_token(&line) != NOATTRIB) break;
		if(!line) break;
		if((bg = attr_token(&line)) < 0 || bg >= COLOURS) break;
	    }
	    if(line) break;
	}
	return ATTRCODE(t, bg);
    }
    printf("#invalid attribute syntax\n");
    printf("Valid options are:\tbold, underline, inverse\n");
    printf("\tor:\t\t[foreground] [on background]\nColours:");
    for(t = 0; t < COLOURS; t++)
	printf(" %s%c", colournames[t], (t == COLOURS - 1) ? '\n' : ',');
    return -1;
}

/*
 * return token number for attribute word or -1 for unrecognized word.
 */
static int attr_token(src)
char **src;
{
    int i;
    char *word = *src;
    *src = chop_str(word, ' ');
    for(i = 0; i < COLOURS; i++)
	if(!strcmp(word, colournames[i])) return i;
    if(!strcmp(word, "bold"))
	return BOLD;
    else if(!strcmp(word, "underline"))
	return UNDERLINE;
    else if(!strcmp(word, "inverse") || !strcmp(word, "reverse"))
	return INVERSE;
    else if(!strcmp(word, "on"))
	return NOATTRIB;	/* fill word */
    return -1;
}

static void cmd_prefix(arg)
char *arg;
{
    if(*arg)
	str_assign(&prefixstr, arg);
    else if(!prefixstr)
	printf("#prefix not set.\n");
    else {
	free(prefixstr); prefixstr = 0;
	if(verbose) printf("#prefix cleared.\n");
    }
}

static void cmd_quote(arg)
char *arg;
{
    /*
     * TODO: "#quote <file". like verbatim mode, but after each enter
     * (or down key on the bottom of the history) a new input line is
     * read from the file.
     */
    toggle(&verbatim, arg);
    printf("#%s mode.\n", verbatim ? "verbatim" : "normal");
}

static void cmd_read(arg)
char *arg;
{
    FILE *fp;
    char buf[LINELEN];
    if(*arg) {
	fp = (arg[0] == '|') ? popen(arg + 1, "r") : fopen(arg, "r");
	if(fp) {
	    while(fgets(buf, LINELEN, fp))
		send_to_host(buf, strlen(buf));
	    if(arg[0] == '|') pclose(fp); else fclose(fp);
	} else
	    printf("#error opening %s\n", arg);
    } else
	printf("#syntax: #read filename or #read |command\n");
}

static void cmd_record(arg)
char *arg;
{
    if(*arg) {
	if(recording_name) {
	    printf("#already recording to alias '%s'\n", recording_name);
	    return;
	}
	chop_arg(arg, ' ');	/* strip any accidental trailing blanks */
	str_assign(&recording_name, arg);
	if(verbose)
	    printf("#recording into alias '%s'\n", recording_name);
    } else if(recording_name) {
	if(verbose)
	    printf("#recording of alias '%s' stopped.\n", recording_name);
	free(recording_name);
	recording_name = 0;
	new_settings = 1;
    } else
	printf("#not recording any alias\n");
}

static void cmd_retrace(arg)
char *arg;
{
    retrace_steps(atoi(arg));
}

#ifndef OMIT_SCHEDULE
static void cmd_schedule(arg)
char *arg;
{
    if(!*arg)
	show_schedule();
    else
	parse_schedule(arg);
}

/*
 * show defined scheduled commands
 */
static void show_schedule()
{
    long delay = 0;
    schedulenode *p;
  
    sprintf(inserted_next, "#%s command%s scheduled%c",
	    schedule ? "The following" : "No",
	    (schedule && !schedule->next) ? " is" : "s are",
	    schedule ? ':' : '.');
    for(p = schedule; paged_print(inserted_next), p; p = p->next) {
	int t = (delay += p->delta) / SCH_HZ;
	sprintf(inserted_next, "#schedule %s\t%d:%2d:%2d=%s", p->label,
		t / 3600, t / 60 % 60, t % 60, p->commands);
    }
}

/*
 * parse the #schedule command
 */
static void parse_schedule(str)
char *str;
{
    long delay = 0;
    schedulenode **np;
    char *rest = chop_arg(str, '=');
    char *k = chop_arg(str, ' ');
    np = lookup_schedule(str);
    if(!rest || !*rest) {
	if(*np) {
	    if(rest) {
		if(verbose)
		    printf("#scheduled command deleted: %s=%s\n",
			    str, (*np)->commands);
		delete_schedulenode(np);
	    } else if(!k) {
		sprintf(inserted_next, "#schedule %s=%s", str, (*np)->commands);
	    } else if((delay = parse_delay(k))) {
		np = reschedule(np, delay);
		if(verbose)
		    printf("#command rescheduled: %s %s=%s\n",
			    str, k, (*np)->commands);
	    }
	} else
	    printf("#no scheduled command with label '%s'\n", str);
    } else if(!*np) {
	static int last = 0;
	static char lbuf[16];
	if((delay = parse_delay(k ? k : str))) {
	    if(!k) {
		do sprintf(lbuf, "_%d", ++last); while(*lookup_schedule(lbuf));
		str = lbuf;
	    }
	    np = add_schedulenode(str, rest, delay);
	    if(verbose)
		printf("#new command scheduled: %s=%s\n", str, rest);
	}
    } else if(!k || (delay = parse_delay(k))) {
	if(strcmp((*np)->commands, rest))
	    str_assign(&(*np)->commands, rest);
	else if(!delay)
	    return;
	if(delay) reschedule(np, delay);
	if(verbose)
	    printf("#scheduled command changed: %s=%s\n", str, rest);
    }
}

/*
 * parse the command delay specification: [[hh:]mm:]ss
 * return the delay in units of 1/SCH_HZ of a second; 0 signifies error.
 */
static long parse_delay(str)
char *str;
{
    char *s = str;
    long r = 0;
    int sign = 0;
    if(*s == '+') s++, sign = 1;
    for( ; ; ) {
	if(*s) {
	    if(isdigit(*s)) r += strtol(s, &str, 10), s = str;
	    else if(*s != ':') break;
	    else s++, r *= 60;
	} else if(r) {
	    r *= SCH_HZ;
	    return sign ? -r : r;
	} else
	    break;
    }
    printf("#schedule: invalid delay specification\n");
    return 0;
}
#endif

static void cmd_set(arg)
char *arg;
{
    /*
     * NOTE: perhaps we should keep certain buffers in our environment,
     * allowing user to directly set them, e.g. "#set FOOTNOTE=HELLO!"
     */
    char *s;
    if(!*arg)
	show_variables();
    else if((s = strchr(arg, '=')) && (*s = 0, !*++s))
	set_env(arg, (char *)0);
    else
	set_env(arg, s ? s : "");
}        

static void show_variables()
{
    int i;
    for(i = 0; environ[i]; i++)
	paged_print(environ[i]);
}

#ifndef NOSHELL
static void cmd_shell(arg)
char *arg;
{
    if(*arg) {
	reset_terminal();
	execute(arg);
	set_terminal();
	lastline();
    }
}
#endif

static void cmd_speedwalk(arg)
char *arg;
{
    if(toggle(&speedwalk, arg) || !parse_walk(arg))
	printf("#speedwalk is now o%s.\n", speedwalk ? "n" : "ff");
}

#ifdef TELNETBUG
static void cmd_stdcolour(arg)
char *arg;
{
    int attrcode = parse_attributes(arg);
    if(attrcode >= 0) {
	int bg = BACKGROUND(attrcode), fg = FOREGROUND(attrcode);
	if(fg >= COLOURS || bg >= COLOURS) {
	    printf("#please specify both foreground and background colours\n");
	} else {
	    sprintf(modenorm, "\033[;3%d;4%dm", fg, bg);
	    printf("%s#standard colour set.\n", modenorm);
	    if(edattrend[0]) strcpy(edattrend, modenorm);
	}
    }
}
#endif

/*ARGSUSED*/
static void cmd_ver(arg)
char *arg;
{
    printver();
}

static void cmd_wrap(arg)
char *arg;
{
    toggle(&wrap, arg);
    printf("#word wrap %sabled.\n", wrap ? "en" : "dis");
}

#ifndef OMIT_SESSIONS
static void cmd_connect(arg)
char *arg;
{
    if(!*arg)
	show_connections();
    else {
	char *host = chop_arg(arg, '=');
	char *init = chop_arg(arg, ' ');
	if(find_sess(arg, 0)) {
	    printf("#session with that name already exists.\n");
	} else if(host) {
	    char *port = chop_arg(host, ' ');
	    connect_sess(arg, init, host, port ? atoi(port) : 0);
	} else
	    connect_sess(arg, init, (char *)0, 0);
    }
}

static void cmd_snoop(arg)
char *arg;
{
    if (!*arg)
	printf("#snoop: no session name\n");
    else
	snoop_sess(arg);
}

static void cmd_zap(arg)
char *arg;
{
    if (!*arg)
	printf("#zap: no session name\n");
    else
	disconnect_sess(arg);
}
#endif

/*
 * if arg is either "on" or "off", toggle a variable and return 1.
 * otherwise return 0.
 */
static int toggle(var, arg)
int *var;
char *arg;
{
    if(!strcmp(arg, "on"))
	*var = 1;
    else if(!strcmp(arg, "off"))
	*var = 0;
    else
	return 0;
    return 1;
}
