/*
 * edit.c
 * line editing functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "cancan.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"

int strcasecmp();
int strncasecmp();

static void prev_line(), next_line();
static void prev_char(), next_char();
static void prev_word(), next_word();
static void del_char_right(), del_char_left();
static void del_word_right(), del_word_left();
static void end_of_line(), beginning_of_line();
static void kill_to_eol(), redraw_input_line();
static void transpose_chars(), transpose_words();
static void enter_into_history(), enter_line();
static void complete_word(), complete_line();
static void matching_bracket();

static void put_in_history();
static void swap_history();
static void demote_word();
static void overtype();
static int skip_word();
static int skip_nonword();

/* circular word completion list */
#define WORD_UNIQUE	1		/* unique among preceding words */
#define WORD_RETAIN	2		/* permanent entry (a #command) */
static int wordindex;
static char *words[WORDSKEPT];
static char wordflags[WORDSKEPT];
#if (WORDSKEPT - 1) >> 8
static unsigned int wordnext[WORDSKEPT];
static unsigned int wordprev[WORDSKEPT];
#else
static unsigned char wordnext[WORDSKEPT];
static unsigned char wordprev[WORDSKEPT];
#endif

/* history buffer */
static char *hist[MAXHIST];	/* saved history lines */
static int curline;		/* current history line */
static int pickline;		/* line to pick history from */

const struct defbind_node defbinds[] = {
  { "enter",	"\n",		0,			enter_line },
  { "return",	"\r",		0,			enter_line },
  { "bs",	"\b",		"&del-char-left",	del_char_left },
  { "del",	"\177",		"&del-char-left",	del_char_left },
  { "tab",	"\t",		"&complete-word",	complete_word },
  { "left",	key_left,	"&prev-char",		prev_char },
  { "right",	key_right,	"&next-char",		next_char },
  { "up",	key_up,		"&prev-line",		prev_line },
  { "down",	key_down,	"&next-line",		next_line },
  { "C-a",	"\001",		"&beginning-of-line",	beginning_of_line },
  { "C-b",	"\002",		"&prev-char",		prev_char },
  { "C-c",	"\003",		0,			intrsig },
  { "C-d",	"\004",		"&del-char-right",	del_char_right },
  { "C-e",	"\005",		"&end-of-line",		end_of_line },
  { "C-f",	"\006",		"&next-char",		next_char },
  { "C-g",	"\007",		"&matching-bracket",	matching_bracket },
  { "C-k",	"\013",		"&kill-to-eol",		kill_to_eol },
  { "C-l",	"\014",		"&redraw-input-line",	redraw_input_line },
  { "C-n",	"\016",		"&next-line",		next_line },
  { "C-p",	"\020",		"&prev-line",		prev_line },
  { "C-t",	"\024",		"&transpose-chars",	transpose_chars },
  { "C-y",	"\031",		"&enter-into-history",	enter_into_history },
  { "C-z",	"\032",		0,			suspend_cancan },
  { "M-tab",	"\033\t",	"&complete-line",	complete_line },
  { "M-b",	"\033b",	"&prev-word",		prev_word },
  { "M-d",	"\033d",	"&del-word-right",	del_word_right },
  { "M-f",	"\033f",	"&next-word",		next_word },
  { "M-t",	"\033t",	"&transpose-words",	transpose_words },
  { "M-del",	"\033\177",	"&del-word-left",	del_word_left },
  { "",		0,		0,			0 }
};

/*
 * redraw the input line
 */
static void redraw_input_line()
{
    clearinpline();
    redraw_input();
}

/*
 * transpose two characters to the left
 */
static void transpose_chars()
{
    if(edlen > 1) {
	char c;
	int k = pos;
	if(k && k-- >= edlen) k--;
	moveto(k);
	c = edbuf[k];
	overtype(edbuf[k + 1]);
	overtype(c);
    }
}

/*
 * transpose current word and the previous one
 */
static void transpose_words()
{
    int a, b, c;
    prev_word();
    a = skip_word(skip_nonword(pos));
    b = skip_nonword(a);
    c = skip_word(b);
    if(b < c) {
	char buf[LINELEN];
	sprintf(buf, "%.*s%.*s%.*s",
		c - b, edbuf + b, b - a, edbuf + a, a - pos, edbuf + pos);
	for(a = 0; buf[a]; overtype(buf[a++]));
    }
}

/*
 * erase everything to the end of line
 */
static void kill_to_eol()
{
    int n = edlen - pos;
    if(n) deletechars(n);
}

/*
 * move cursor to end of line
 */
static void end_of_line()
{
    moveto(edlen);
}

/*
 * move cursor to beginning of line
 */
static void beginning_of_line()
{
    moveto(0);
}

/*
 * delete a character to the right
 */
static void del_char_right()
{
    if(pos < edlen)
	deletechars(1);
}

/*
 * delete a character to the left
 */
static void del_char_left()
{
    if(pos) {
	moveto(pos - 1);
	deletechars(1);
    }
}

/*
 * move the editline into history
 */
static void enter_into_history()
{
    if(edlen) {
	clearinpline();
	put_in_history(edbuf);
	pickline = curline;
	edbuf[0] = '\0';
	pos = edlen = 0;
    }
}

/*
 * enter a line
 */
static void enter_line()
{
    int ls = (pos >= edlen) ? 0 : CURLINE(edlen) - CURLINE(pos);
    do newline(); while(ls--); 
    if(capfile) fprintf(capfile, "%s\n", edbuf);
#ifdef NEWBIE
    denewbiefy();
#endif
    inserted_next[0] = '\0';
    if(verbatim && edbuf[0] != '#' && edbuf[0] != '{') {
	/* BUG: we should process '\' as a first char in verbatim mode */
	edbuf[edlen] = '\n';
	send_to_host(edbuf, edlen + 1);
	edbuf[edlen] = '\0';
    } else {
	hilite(0);
	new_settings = 0;
        verbose = 1;
	parse(edbuf, 1, (struct param *)0);
	verbose = 0;
	if(new_settings)	/* user defined something -- save */
	    save_settings();
    }
    if(edlen > 1) {
	/* don't put identical lines in history, nor short ones */
	char *p = hist[curline ? curline - 1 : MAXHIST - 1];
	if(!p || strcmp(edbuf, p)) put_in_history(edbuf);
    }
    pickline = curline;
    edbuf[0] = '\0';
    if(!pager_text) {
	if(inserted_next[0])
	    escape_command(inserted_next, edbuf);
	else if(prefixstr)
	    strcpy(edbuf, prefixstr);
    }
    pos = edlen = strlen(edbuf);
    redraw_prompt();
    redraw_input();
}

/*
 * move one word forward
 */
static void next_word()
{
    moveto(skip_word(skip_nonword(pos)));
}

/*
 * move one word backward
 */
static void prev_word()
{
    int i = pos;
    while(i && !islabel(edbuf[i - 1])) i--;
    while(i && islabel(edbuf[i - 1])) i--;
    moveto(i);
}

/*
 * delete word to the right
 */
static void del_word_right()
{
    int n = skip_word(skip_nonword(pos)) - pos;
    if(n) deletechars(n);
}

/*
 * delete word to the left
 */
static void del_word_left()
{
    int i = pos;
    prev_word();
    deletechars(i - pos);
}

/*
 * get previous line from history list
 */
static void prev_line()
{
    int i = pickline ? pickline - 1 : MAXHIST - 1;
    if(i != curline)
	swap_history(i);
}

/*
 * get next line from history list
 */
static void next_line()
{
    if(pickline != curline) {
	int i = pickline + 1;
	if(i >= MAXHIST) i = 0;
	swap_history(i);
    }
}

/*
 * move one char backward
 */
static void prev_char()
{
    if(pos)
	moveto(pos - 1);
}

/*
 * move one char forward
 */
static void next_char()
{
    if(pos < edlen)
	moveto(pos + 1);
}
 
/*
 * move to matching bracket
 */
static void matching_bracket()
{
    char *p, *brackets = "(){}[]<>";
    int i = pos;
    if((p = strchr(brackets, edbuf[i]))) {
	int nested = 0;
	int dir = (p - brackets) & 1 ? -1 : 1;
	while((i += dir) >= 0 && i < edlen)
	    if(edbuf[i] == p[0]) nested++;
	    else if(edbuf[i] == p[dir] && !nested--) {
		moveto(i);
		break;
	    }
    }
}

/*
 * match and complete a word referring to the word list
 */
static void complete_word()
{
    static int curr_word, comp_len, root_len;
    char *root, *p;
    int k, n;

    /* determine curr_word, root, root_len */
    if(last_edit_func == complete_word && (k = comp_len)) {
	moveto(pos - k);
	n = pos - root_len;
    } else {
#ifdef DELIM
	for(n = pos; n && !strchr(DELIM, edbuf[n - 1]); n--);
#else
	for(n = pos; n && islabel(edbuf[n - 1]); n--);
#endif
	k = 0;  curr_word = wordindex;  root_len = pos - n;
    }
    root = edbuf + n;  comp_len = 0;

    /* look up next valid completion and print it */
    while((p = words[curr_word = wordnext[curr_word]]))
	if(!strncasecmp(p, root, root_len) && *(p += root_len) &&
		(n = strlen(p)) + edlen < LINELEN) {
	    comp_len = n;
	    while(k ? (k--, overtype(*p++), --n) : (insertchars(p, n), 0));
	    break;
	}
    if(k) deletechars(k);

    /* delete duplicate instances of the word */
    if(p && !(wordflags[k = curr_word] & WORD_UNIQUE)) {
	wordflags[k] += WORD_UNIQUE;
	p = words[k];  n = wordnext[k];
	while(words[k = n])
	    /*
	     * NOTE: could use strcmp() here, and change the code above
	     * to overtype the whole word, including the root of it.
	     */
	    if(n = wordnext[k], !strcasecmp(p, words[k])) {
		demote_word(k);
		free(words[k]);  words[k] = 0;
		wordflags[curr_word] |= wordflags[k];	/* move retain flag */
		if((wordflags[k] &= WORD_UNIQUE)) break;
	    }
    }
}

/*
 * match backwards in history and complete entire lines
 */
static void complete_line()
{
    if(edlen) {
	int i = curline;
	for( ; ; ) {
	    if(!i) i = MAXHIST;
	    if(--i == curline) break;
	    if(hist[i] && !strncmp(hist[i], edbuf, edlen)) {
		clearinpline();
		strcpy(edbuf, hist[i]);
		pos = edlen = strlen(edbuf);
		redraw_input();
		break;
	    }
	}
    }
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
static void put_in_history(str)
char *str;
{
    int i, c;
    str_assign(hist + curline, str);
    if(++curline >= MAXHIST) curline = 0;
  
    /* split into words and put into completion list (unless too short) */
    do {
#ifdef DELIM
	str += strspn(str, DELIM);
	if(!*str) break;
	i = strcspn(str, DELIM);
#else
	for( ; ; str++)
	    if(!*str) return;
	    else if(islabel(*str)) break;
	for(i = 0; islabel(str[i]); i++);
#endif
	c = str[i];  str[i] = 0;
	if(i >= 3) put_word(str);
        str += i + 1;
    } while(c);
}

/*
 * swap editline with a history line
 */
static void swap_history(line)
int line;
{
    if(hist[line]) {
	str_assign(hist + pickline, edbuf);
	pickline = line;
	clearinpline();
	strcpy(edbuf, hist[line]);
	pos = edlen = strlen(edbuf);
	redraw_input();
    }
}

/*
 * put word in word completion ring
 */
void put_word(s)
char *s;
{
    int r = wordindex;
    words[r] = str_dup(s);  wordflags[r] = 0;
    while(wordflags[r = wordprev[r]] & WORD_RETAIN);
    demote_word(r);  wordindex = r;
    if(words[r]) { free(words[r]);  words[r] = 0; }
}

/*
 * reinsert a node just before wordindex ie make it the last word
 */
static void demote_word(i)
int i;
{
    wordnext[wordprev[i]] = wordnext[i];
    wordprev[wordnext[i]] = wordprev[i];
    wordprev[i] = wordprev[wordnext[i] = wordindex];
    wordprev[wordindex] = wordnext[wordprev[wordindex]] = i;
}

/*
 * print a char on current position (overwrite), advance position
 */
static void overtype(c)
char c;
{
    edbuf[pos++] = c;
    putchar(c);
    if(!CURCOL(pos))
	putchar('\n');
}

/*
 * return the position after leading islabel() chars
 */
static int skip_word(i)
int i;
{
    while(islabel(edbuf[i])) i++;
    return i;
}

/*
 * return the position after leading word separator chars
 */
static int skip_nonword(i)
int i;
{
    while(i < edlen && !islabel(edbuf[i])) i++;
    return i;
}

/*
 * initialize the word completion ring
 */
void default_completions()
{
    char buf[LINELEN];
    const cmdstruct *p;
    int i;
    /* NOTE: perhaps keep the '#' command prefix in commands struct? */
    for(i = 0, buf[0] = '#', p = commands; p->name; p++)
	if(p->funct && strlen(p->name) >= 3) {
	    if(++i >= WORDSKEPT) break;
	    strcpy(buf + 1, p->name);
	    words[i] = str_dup(buf);
	    wordflags[i] = WORD_UNIQUE | WORD_RETAIN;
	}
    wordprev[0] = WORDSKEPT - 1;  wordnext[WORDSKEPT - 1] = 0;
    for(i = WORDSKEPT - 1; i; wordnext[i - 1] = i, wordprev[i] = i - 1, i--);
}

/*
 * execute string as if typed
 */
void run_command(cmd)
char *cmd;
{
    if(pager_text) pager(0);
    else clearinpline();
    hilite(1);
    printf("%s\n", cmd);
    hilite(0);
    if(capfile) fprintf(capfile, "%s\n", cmd);
    parse(cmd, 1, (struct param *)0);
    redraw_prompt();
    redraw_input();
}

