/*
 * list.c - list utility functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "cancan.h"
#include "list.h"

aliasnode *aliases;	/* head of alias list */
actionnode *actions;	/* head of action list */
marknode *markers;	/* head of marker list */
keynode *keydefs;	/* head of key binding list */
#ifndef OMIT_SCHEDULE
schedulenode *schedule;	/* head of scheduled commands list */
#endif

/*
 * generic list node adding routine
 */
defnode **add_node(newnode, base)
defnode *newnode, **base;
{
#ifndef NOSORT
    while((*base) && strcmp((*base)->field, newnode->field) < 0)
	base = &(*base)->next;
#endif
    newnode->next = *base;
    *base = newnode;
    return base;
}

/*
 * generic list node lookup routine
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
defnode **lookup(field, base)
char *field;
defnode **base;
{
    register defnode **p = base;
    while(*p && strcmp(field, (*p)->field))
	p = &(*p)->next;
    return p;
}


/*
 * add a node to the alias list (sorted unless NOSORT is defined)
 */
aliasnode **add_aliasnode(name, subst)
char *name, *subst;
{
    aliasnode *new = (aliasnode*)malloc(sizeof(aliasnode));
    if(!new) syserr("malloc");
    new->name = str_dup(name);
    new->subst = str_dup(subst);
    return (aliasnode **)add_node((defnode*)new, (defnode**)&aliases);
}

#ifndef lookup_alias
/*
 * look up an alias node by name:
 */
aliasnode **lookup_alias(name)
char *name;
{
    aliasnode **p = &aliases;
    while(*p && strcmp(name, (*p)->name))
	p = &(*p)->next;
    return p;
}
#endif

/*
 * delete an alias node, given a pointer to its predecessor's pointer
 */
void delete_aliasnode(base)
aliasnode **base;
{
    aliasnode *p = *base;
    *base = p->next;
    free(p->subst);
    free(p->name);
    free((char *)p);
}


/*
 * add a node to the action list (sorted unless NOSORT is defined)
 */
actionnode **add_actionnode(pattern, commands, label, active)
char *pattern, *commands, *label; int active;
{
    actionnode *new = (actionnode*)malloc(sizeof(actionnode));
    if(!new) syserr("malloc");
    new->pattern = str_dup(pattern);
    new->commands = str_dup(commands);
    new->label = str_dup(label);
    new->active = active;
    return (actionnode **)add_node((defnode*)new, (defnode**)&actions);
}

#ifndef lookup_action
/*
 * look up an action node by label:
 */
actionnode **lookup_action(label)
char *label;
{
    actionnode **p = &actions;
    while(*p && strcmp(label, (*p)->label))
	p = &(*p)->next;
    return p;
}
#endif

#ifndef lookup_action_pattern
/*
 * look up an action node by pattern:
 */
actionnode **lookup_action_pattern(pattern)
char *pattern;
{
    actionnode **p = &actions;
    while(*p && strcmp(pattern, (*p)->pattern))
	p = &(*p)->next;
    return p;
}
#endif

/*
 * delete an action node, given a pointer to its predecessor's pointer
 */
void delete_actionnode(base)
actionnode **base;
{
    actionnode *p = *base;
    *base = p->next;
    free(p->label);
    free(p->commands);
    free(p->pattern);
    free((char*)p);
}


/*
 * add a node to the marker list (sorted unless NOSORT is defined)
 */
marknode **add_marknode(pattern, attrcode)
char *pattern; int attrcode;
{
    marknode *new = (marknode*)malloc(sizeof(marknode));
    if(!new) syserr("malloc");
    new->pattern = str_dup(pattern);
    new->plen = strlen(new->pattern);
    new->attrcode = attrcode;
    return (marknode **)add_node((defnode*)new, (defnode**)&markers);
}

#ifndef lookup_marker
/*
 * look up a marker node by pattern:
 */
marknode **lookup_marker(pattern)
char *pattern;
{
    marknode **p = &markers;
    while(*p && strcmp(pattern, (*p)->pattern))
	p = &(*p)->next;
    return p;
}
#endif

/*
 * delete an marker node, given a pointer to its predecessor's pointer
 */
void delete_marknode(base)
marknode **base;
{
    marknode *p = *base;
    *base = p->next;
    free(p->pattern);
    free((char*)p);
}


/*
 * add a node to the keydef list (sorted unless NOSORT is defined)
 */
keynode **add_keynode(name, sequence, funct, call_data)
char *name, *sequence; void (*funct)(); char *call_data;
{
    keynode *new = (keynode*)malloc(sizeof(keynode));
    if(!new) syserr("malloc");
    new->name = str_dup(name);
    new->sequence = str_dup(sequence);
    new->funct = funct;
    if(call_data) call_data = str_dup(call_data);
    new->call_data = call_data;
    return (keynode **)add_node((defnode*)new, (defnode**)&keydefs);
}

#ifndef lookup_key
/*
 * look up a keydef node by pattern:
 */
keynode **lookup_key(name)
char *name;
{
    keynode **p = &keydefs;
    while(*p && strcmp(name, (*p)->name))
	p = &(*p)->next;
    return p;
}
#endif

/*
 * delete a keydef node, given a pointer to its predecessor's pointer
 */
void delete_keynode(base)
keynode **base;
{
    keynode *p = *base;
    *base = p->next;
    free(p->sequence);
    free(p->name);
    if(p->call_data) free(p->call_data);
    free((char*)p);
}

#ifndef OMIT_SCHEDULE
/*
 * add a node to the schedule list (sorted by time)
 */
schedulenode **add_schedulenode(label, commands, delay)
char *label, *commands; long delay;
{
    schedulenode *new = (schedulenode*)malloc(sizeof(schedulenode));
    if(!new) syserr("malloc");
    new->label = str_dup(label);
    new->commands = str_dup(commands);
    new->delta = 0;
    new->next = schedule;
    schedule = new;
    return reschedule(&schedule, delay);
}

/*
 * reinsert a schedulenode with new delay, keeping the list sorted
 * minus sign of delay indicates that the current delay is increased
 * rather than set to a new value
 */
schedulenode **reschedule(np, delay)
schedulenode **np;  long delay;
{
    schedulenode *old = *np;
    if((*np = old->next))
	old->next->delta += old->delta;
    if(delay > 0)
	np = &schedule;
    else
	delay = (old->delta > 0 ? old->delta : 0) - delay;
    for( ; *np && (*np)->delta <= delay; np = &(*np)->next)
	delay -= (*np)->delta;
    old->delta = delay;
    if((old->next = *np))
	old->next->delta -= delay;
    *np = old;
    return np;
}

#ifndef lookup_schedule
/*
 * look up a scheduled command by label:
 * return pointer to pointer to node or a pointer to NULL if nothing found
 */
schedulenode **lookup_schedule(label)
char *label;
{
    schedulenode **p = &schedule;
    while(*p && strcmp(label, (*p)->label))
	p = &(*p)->next;
    return p;
}
#endif

/*
 * delete an schedule node, given a pointer to its predecessor's pointer
 */
void delete_schedulenode(base)
schedulenode **base;
{
    schedulenode *p = *base;
    if((*base = p->next))
	p->next->delta += p->delta;
    free(p->label);
    free(p->commands);
    free((char*)p);
}
#endif
