/*
 * term.c - terminal handling routines for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/ioctl.h>

#ifndef USESGTTY
#ifdef APOLLO
#include "/sys5.3/usr/include/sys/termio.h"
#else
#include <termio.h>
#include <termios.h>
#endif
#else /* USESGTTY */
#include <bsd/sgtty.h>
#endif

#ifdef BSD_386
#include <sys/ioctl_compat.h>
#define O_RAW RAW
#define O_ECHO ECHO
#endif

#ifndef USESGTTY
/*
 * NOTE: should probably use TCSETSW/TCSETAW/TIOCSETAW instead,
 * since we could be toggling the OPOST flag...
 */
#if defined(TCSETS) || defined(TCSETATTR)
#ifndef TCSETS		/* HP-UX defines TCSETATTR/TCGETATTR? */
#define TCSETS TCSETATTR
#define TCGETS TCGETATTR
#endif
typedef struct termios termiostruct;
#else
/* if TCSETA is not defined either, we could try to use TIOCSETA */
#define TCSETS TCSETA
#define TCGETS TCGETA
typedef struct termio termiostruct;
#endif
#endif /* !USESGTTY */

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"

static char *str_mark();
static void strins();
static void gotoxy();
static void optmove();
static int pager_done();
static int pager_prompt();

#ifndef VT100

int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();

static int extract();

/* terminal escape sequences */
char kpadstart[CAPLEN], kpadend[CAPLEN];
char begoln[CAPLEN], clreoln[CAPLEN], clreoscr[CAPLEN];
char leftcur[CAPLEN], upcur[CAPLEN], curgoto[CAPLEN];
char delchar[CAPLEN], inschar[CAPLEN], insstop[CAPLEN];
/* attribute changers: */
char modebold[CAPLEN], modeinv[CAPLEN], modeuline[CAPLEN];
char modestandon[CAPLEN], modestandoff[CAPLEN], modenorm[CAPLEN];

char key_left[CAPLEN], key_right[CAPLEN];
char key_up[CAPLEN], key_down[CAPLEN];

static int len_begoln, len_clreoln;
static int len_leftcur, len_upcur, gotocost;
static int deletecost, insertcost, insertfinish;

int metakey;		/* terminal has meta key setting high bit */
int wrapglitch;		/* strange wrap behaviour (a la vt100) */

#else	/* hard-coded vt100 features if no termcap: */

char kpadstart[] = "", kpadend[] = "";
char begoln[] = "\r";
char clreoln[] = "\033[K", clreoscr[] = "\033[J";
char leftcur[] = "\033[D", upcur[] = "\033[A";
char modebold[] = "\033[1m", modeinv[] = "\033[7m", modeuline[] = "\033[4m";
char modestandon[] = "", modestandoff[] = "", modenorm[CAPLEN] = "\033[m";

char key_left[] = "\033[D", key_right[] = "\033[C";
char key_up[] = "\033[A", key_down[] = "\033[B";

#define len_begoln	1
#define len_leftcur	3
#define len_upcur	3
#define gotocost	8

#define wrapglitch	1
#define insertfinish	0

#endif /* !VT100 */

#ifndef LINES
#define LINES 24
#endif
#ifndef COLUMNS
#define COLUMNS 80
#endif

int lines = LINES;	/* screen size */
int cols = COLUMNS;

int lines_1 = LINES - 1;/* for efficiency's sake */
int cols_1 = COLUMNS;	/* either cols, or cols-1 if no wrapglitch */

int line0;		/* screen line where the input line starts */
int col0;		/* input line offset (= length of prompt) */

/* input line highlighting */
int edattr;		/* input line attributes defined */
char edattrbeg[CAPLEN];	/* start input line attributes */
char edattrend[CAPLEN]; /* end input line attributes */

/* pager data */
int pager_text;
static defnode *pager_head, *pager_tail;

#ifndef USESGTTY
static termiostruct ttybsave;
#else
static struct sgttyb ttybsave;
static struct tchars tcsave;
static struct ltchars ltcsave;
#endif

/*
 * Save the original state, disable echo and special keys.
 */
void set_terminal()
{
#ifndef USESGTTY
    ioctl(0, TCGETS, &ttybsave);
#else
    ioctl(0, TIOCGETP, &ttybsave);
    ioctl(0, TIOCGETC, &tcsave);
    ioctl(0, TIOCGLTC, &ltcsave);
#endif
    set_spec_keys(0);
    signal(SIGINT, intrsig);
    fputs(kpadstart, stdout);
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
    signal(SIGINT, SIG_IGN);
#ifndef USESGTTY
    ioctl(0, TCSETS, &ttybsave);
#else
    ioctl(0, TIOCSETP, &ttybsave);
    ioctl(0, TIOCSETC, &tcsave);
    ioctl(0, TIOCSLTC, &ltcsave);
#endif
    hilite(0);
    fputs(kpadend, stdout);
}

/*
 * set raw/cooked mode depending on the current linemode
 * enable/disable special keys
 */
void set_spec_keys(k)
int k;
{
#ifndef USESGTTY
    termiostruct ttyb;
    ttyb = ttybsave;
    ttyb.c_cc[VTIME] = 0;
    ttyb.c_cc[VMIN] = 1;
    ttyb.c_oflag |= OPOST;
    if(linemode & LM_CHAR) ttyb.c_oflag &= ~OPOST;
    ttyb.c_lflag &= ~(ECHO | ISIG | ICANON | IEXTEN);
    if(k) ttyb.c_lflag |= ISIG;
    ioctl(0, TCSETS, &ttyb);
#else /* USESGTTY */
    static const struct tchars tc_off = { -1, -1, -1, -1, -1, -1 };
    static const struct ltchars ltc_off = { -1, -1, -1, -1, -1, -1 };
    const struct tchars *tc = &tc_off;
    const struct ltchars *ltc = &ltc_off;
    struct sgttyb ttyb;
    ttyb = ttybsave;
    ttyb.sg_flags &= ~(O_RAW | O_ECHO);
    ttyb.sg_flags |= O_CBREAK | O_CRMOD;
    if(linemode & LM_CHAR) ttyb.sg_flags ^= O_RAW | O_CRMOD;
    ioctl(0, TIOCSETP, &ttyb);
    if(k) tc = &tcsave, ltc = &ltcsave;
    ioctl(0, TIOCSETC, tc);
    ioctl(0, TIOCSLTC, ltc);
#endif
}

/*
 * read termcap definitions
 */
void settermcap()
{
#ifndef VT100
    struct tc_init_node {
	char cap[3], critical;
	char *buf;
	int *len;
    };
    static const struct tc_init_node tc_init[] = {
	{ "cm", 1, curgoto, 0 },
	{ "ce", 1, clreoln, &len_clreoln },
	{ "cd", 1, clreoscr, 0 },
#if 0	/* we don't use it anywhere... */
	{ "nd", 0, rightcur, 0 },
#endif
	{ "le", 0, leftcur, &len_leftcur },
	{ "up", 0, upcur, &len_upcur },
	{ "cr", 0, begoln, &len_begoln },
	{ "im", 0, inschar, &insertcost },
	{ "ic", 0, inschar, &insertcost },
	{ "ei", 0, insstop, &insertcost },
	{ "dm", 0, delchar, &deletecost },
	{ "dc", 0, delchar, &deletecost },
	{ "ed", 0, delchar, &deletecost },
	{ "me", 0, modenorm, 0 },
	{ "md", 0, modebold, 0 },
	{ "mr", 0, modeinv, 0 },
	{ "us", 0, modeuline, 0 },
	{ "so", 0, modestandon, 0 },
	{ "se", 0, modestandoff, 0 },
	{ "ks", 0, kpadstart, 0 },
	{ "ke", 0, kpadend, 0 },
	{ "kl", 0, key_left, 0 },
	{ "kr", 0, key_right, 0 },
	{ "ku", 0, key_up, 0 },
	{ "kd", 0, key_down, 0 },
	{ "",   0, 0, 0 }
    };
    const struct tc_init_node *np;
    char tcbuf[2048];		/* by convention, this is enough */
    int i;
#endif    
#if !defined(VT100) || defined(TELNETBUG)
    char *term = getenv("TERM");
#endif
#ifndef VT100
    if(!term || !(i = tgetent(tcbuf, term))) {
	if(term) fprintf(stderr,
		    "There is no entry for '%s' in the terminal data base.\n",
		    term);
	fprintf(stderr, "Please set your TERM environment variable.\n");
	exit(1);
    }
    if(i == -1)	/* NOTE: perhaps we should assume default vt100 values? */
	exit(syswarn("Couldn't read termcap database"));
    for(np = tc_init; np->cap[0]; np++)
	if((i = extract(np->cap, np->buf))) {
	    if(np->len) *np->len += i;
	} else if(np->critical) {
	    fprintf(stderr,
		    "Your terminal is not powerful enough, missing '%s'.\n",
		    np->cap);
	    exit(1);
	}
    if(!len_begoln)
	strcpy(begoln, "\r"), len_begoln = 1;
    if(!len_leftcur)
	strcpy(leftcur, "\b"), len_leftcur = 1;
    /* NOTE: check_winsize() will override these if it can get the term size */
    if((i = tgetnum("li")) > 0) lines_1 = (lines = i) - 1;
    if((i = tgetnum("co")) > 1) cols_1 = cols = i;
    gotocost = strlen(tgoto(curgoto, cols_1, lines_1));
    metakey = tgetflag("km");
    wrapglitch = tgetflag("xn");
    if(!wrapglitch)
	cols_1--, insertfinish = gotocost + len_clreoln;
#endif
#ifdef TELNETBUG
    if(!term || !strncmp(term, "vt10", 4)) {
	/* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
	sprintf(modenorm, "\033[;3%d;4%dm", DEFAULTFG, DEFAULTBG);
    }
#endif
}

/*
 * add the default keypad bindings to the list
 */
void keypad_bindings()
{
    /*
     * NOTE: termcap doesn't have sequences for the numeric keypad,
     * so we just assume they are the same as for a vt100.
     */
    add_keynode("kp2", "\033Or", run_command, "s");
    add_keynode("kp3", "\033Os", run_command, "d");
    add_keynode("kp4", "\033Ot", run_command, "w");
    add_keynode("kp5", "\033Ou", run_command, "exits");
    add_keynode("kp6", "\033Ov", run_command, "e");
    add_keynode("kp8", "\033Ox", run_command, "n");
    add_keynode("kp9", "\033Oy", run_command, "u");
}
    
/*
 * initialize the key binding list
 */
void default_bindings()
{
    /*
     * NOTE: no check is made to see if (up, down, etc.) keys have valid
     * keysequence. could use key_collision(), which is defined in cmd.c
     */
    const struct defbind_node *p = defbinds;
    do {
	if(p->seq[0]) add_keynode(p->label, p->seq, p->funct, (char *)0);
    } while((++p)->label[0]);
}

#ifndef VT100
/*
 * extract termcap 'cap' and strcat it to buf.
 * return the lenght of the extracted string.
 */
static int extract(cap, buf)
char *cap, *buf;
{
    static char *bp;
    char *d = buf + strlen(buf);
    char *s = tgetstr(cap, (bp = d, &bp));
    int len;
    if(!s) return (*bp = 0);
    /*
     * NOTE: Padding information is removed. We assume that no terminals
     * need padding nowadays. At least it makes things much easier.
     */
    s += strspn(s, "0123456789*");
    for(len = 0; *s; *d++ = *s++, len++)
	if(*s == '$' && *(s + 1) == '<')
	    if(!(s = strchr(s, '>')) || !*++s) break;
    *d = 0;
    return len;
}
#endif /* VT100 */

/*
 * position the cursor using absolute coordinates
 * NOTE: tgoto is quite a piece of code, willing to handle padding,
 * which we chose to ignore, etc., etc.
 */
static void gotoxy(col, line)
int col, line;
{
#ifdef VT100
  printf("\033[%d;%dH", line + 1, col + 1);
#else
  fputs(tgoto(curgoto, col, line), stdout);
#endif
}

/*
 * optimized cursor movement
 * from (fromline, fromcol) to (toline, tocol)
 * if tocol > 0, (toline, tocol) must reside on editline.
 */
static void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    static char buf[CAPLEN];
    char *cp = buf;
    int i, cost = 0;
    /*
     * First, move vertically to the correct line, then horisontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use that.
     */
    for(;;) {
	if((i = toline - fromline) < 0) {
	    if(!len_upcur || (cost = -i * len_upcur) >= gotocost) break;
	    do strcpy(cp, upcur), cp += len_upcur; while(++i);
	} else if(i) {		/* lf's are mapped into crlf on output */
	    if((cost = 2 * i) >= gotocost) break;
	    do *cp++ = '\n'; while(--i);
	    fromcol = 0;
	}
	if(fromcol >= cols) break;	/* we are in a glitch state */
	if((i = tocol - fromcol) < 0) {
	    int dist = -i * len_leftcur;
	    if(dist <= len_begoln + tocol) {
		if(cost + dist > gotocost) break;
		do strcpy(cp, leftcur), cp += len_leftcur; while(++i);
	    } else {
		if((cost += len_begoln) >= gotocost) break;
		strcpy(cp, begoln);  cp += len_begoln;
		fromcol = 0;  i = tocol;
	    }
	}
	/*
	 * NOTE: in theory we could use ("ta" hardware) tabs to move the
	 * cursor horizontally. Unless the terminal has "xt" destructive
	 * tabs glitch, of course. Nah...
	 */
	if(i) {
	    /* if hiliting in effect, just use gotoxy */
	    if(cost + i > gotocost || edattr) break;
	    if(fromcol < col0 && toline == line0) {
		/*
		 * FEATURE: it should be recorded somewhere whether printed
		 * prompt was marked. In that case, gotoxy must be used
		 */
		if(promptstr[col0]) break;	/* special chars in prompt */
		strcpy(cp, promptstr);
		cp += fromcol = col0;
	    }
	    memcpy(cp, edbuf + (toline - line0) * cols_1 + fromcol - col0,
		    tocol - fromcol);
	    cp += tocol - fromcol;
	}
	*cp = 0;
	fputs(buf, stdout);
	return;
    }
    gotoxy(tocol, toline);
}

/*
 * change the position on input line (optmove there, and set pos)
 */
void moveto(new_pos)
int new_pos;
{
    /*
     * FEATURE: the line we are moving to might be less than 0, or greater
     * than lines_1, if the display is too small to hold the whole editline.
     * In that case, the input line should be (partially) redrawn.
     */
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(new_pos), CURCOL(new_pos));
    pos = new_pos;
}

/*
 * move to the beginning of last line
 */
void lastline()
{
    /* NOTE: could use tcap "ll" here */
    gotoxy(0, line0 = lines_1);
}

/*
 * display or clear a one-line footnote at the bottom of the screen
 */
void display_footnote(show)
int show;
{
    static int visible;
    if(show || visible) {
	int to_c = CURCOL(pos);
	int to_l = CURLINE(pos);
	optmove(to_l, to_c, lines_1, 0);
	if(visible) fputs(clreoln, stdout);
	if((visible = show)) {
	    hilite(0);
	    if(CURLINE(edlen) >= lines_1) {
		line0--;  to_l--;  putchar('\n');
	    }
	    fputs(footnote, stdout);
	    gotoxy(to_c, to_l);
	} else {
	    footnote[0] = 0;
	    optmove(lines_1, 0, to_l, to_c);
	}
    }
}

/*
 * clear current input line (but not the prompt)
 * cursor is left right after the prompt
 */
void clearinpline()
{
    if(edlen) {
	int from_c = CURCOL(pos);
	int from_l = CURLINE(pos);
	if(line0 < lines_1) {
	    /* NOTE: plain clreoscr would suffice when col0 == 0 */
	    optmove(from_l, from_c, line0 + 1, 0);
	    fputs(clreoscr, stdout);
	    from_c = 0;  from_l = line0 + 1;
	}
	optmove(from_l, from_c, line0, col0);
	fputs(clreoln, stdout);
    }
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    if(edlen && !pager_text) {
	int k = cols_1 - col0;
	hilite(1);
	printf("%.*s", k, edbuf);
	if(edlen >= k) {
	    int n = lines_1;
	    do {
		printf("\n%.*s", cols_1, edbuf + k);
		--n;  k += cols_1;
	    } while(k <= edlen);
	    if(n < line0) line0 = n;
	}
	k = pos;  pos = edlen;
	moveto(k);
    }
}

/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
void deletechars(n)
int n;
{
    int r_cost, d_cost, p = pos;
    int p_col = CURCOL(p);
    int p_line = CURLINE(p);
    int nl = p - p_col;		/* next line's starting pos */
    int cl = p_line;		/* current line */
#if 0
    memmove(edbuf + p, edbuf + p + n, edlen - p - n);
    memset(edbuf + edlen - n, ' ', n);
#else
    for(d_cost = p + n; d_cost < edlen; d_cost++)
	edbuf[d_cost - n] = edbuf[d_cost];
    for(d_cost -= n; d_cost < edlen; d_cost++)
	edbuf[d_cost] = ' ';
#endif
    for( ; ; ) {
	d_cost = 0;
	/* NOTE: redraw might need to goto back whereas delete would not. */
	if((r_cost = edlen) > (nl += cols_1))
	    r_cost = nl, d_cost = n + gotocost;
	r_cost -= p;
#ifndef VT100
	/* NOTE: no clreoln is used here. Simplicity */
	if(deletecost && deletecost * n + d_cost < r_cost) {
	    for(d_cost = n; d_cost; d_cost--) fputs(delchar, stdout);
	    if(edlen <= nl) break;
	    gotoxy(cols_1 - n, cl);
	    printf("%.*s", n, edbuf + nl - n);
	} else
#endif
	{
	    p += printf("%.*s", r_cost, edbuf + p);
	    if(edlen <= nl) break;
	}
	putchar('\n'); p = nl; cl++;
	if(p + n >= edlen) { fputs(clreoscr, stdout); break; }
    }
    edbuf[edlen -= n] = '\0';
    optmove(cl, p - nl + cols_1, p_line, p_col);
}

/*
 * insert one character first in a string
 */
static void strins(s, c)
char *s; char c;
{
  char d;
  for(; *s; d = *s, *s++ = c, c = d);
  *s++ = c; *s = 0;
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
void insertchars(str, n)
char *str; int n;
{
    int r_cost, i_cost = n, p = pos;
    int nl = p - CURCOL(p);	/* next line's starting pos */
    int cl = CURLINE(p);	/* current line */
#if 0
    memmove(edbuf + p + n, edbuf + p, edlen - p);
    memcpy(edbuf + p, str, n);
#else
    while(i_cost) strins(edbuf + p, str[--i_cost]);
#endif
    edlen += n;  pos += n;
    do {
	i_cost = n;
	if((r_cost = edlen) > (nl += cols_1))
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
#ifndef VT100
	/*
	 * NOTE: insert mode is used only when one char is inserted
	 * (which is probaly true > 95% of the time). Simplicity
	 */
	if(n == 1 && insertcost && insertcost + i_cost < r_cost) {
	    printf("%s%c%s", inschar, edbuf[p++], insstop);
	    if(insertfinish && edlen > nl) {
		gotoxy(cols_1, cl);
		fputs(clreoln, stdout);
	    }
	} else
#endif
	{
	    printf("%.*s", r_cost, edbuf + p);  p += r_cost;
	}
	if(edlen < nl) break;
	putchar('\n');  p = nl;
	if(++cl > lines_1) cl = lines_1, line0--;
    } while(edlen > nl);
    optmove(cl, p - nl + cols_1, CURLINE(pos), CURCOL(pos));
}

/*
 * print a newline and advance line0.
 */
void newline()
{
    putchar('\n');
    if(line0 < lines_1) line0++;
}

/*
 * toggle input line attributes
 */
void hilite(k)
{
    static in_effect;
    if(in_effect) {
	if(!k) fputs(edattrend, stdout);
    } else {
	if(k) fputs(edattrbeg, stdout);
    }
    in_effect = k;
}

/*
 * put escape sequences to turn on/off an attribute into given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{	
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
    *begin = *end = '\0';
    if(fg < COLOURS) {
	if(bg < COLOURS)
	    sprintf(begin, "\033[3%d;4%dm", fg, bg);
	else
	    sprintf(begin, "\033[3%dm", fg);
    } else if(bg < COLOURS) {
	sprintf(begin, "\033[4%dm", bg);
    } else {
	char *src;
	if(fg == UNDERLINE) src = modeuline;
	else if(fg == BOLD) src = modebold;
	else if(fg == INVERSE) src = modeinv;
	else return;
	if(!*src) {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	    return;
	}
	strcpy(begin, src);
    }
    if(attrcode != NOATTRCODE)
	strcpy(end, modenorm);
}

/*
 * search for marker in string
 * marker must not be empty string
 *
 * FEATURE: str_mark ought perhaps match a marker starting/ending with a
 * word separator char regardless if it's preceded/followed by one or not.
 */
static char *str_mark(str, mark)
char *str, *mark;
{
#if 1
    register char *s = str;
    register char *m = mark;
    if(*s) for( ; ; ) {
	if(*m++ == *s++)
	    if(*m) continue;
	    else if(!islabel(*s)) return str;
	m = mark;  s = str;
	while(*++s && (*s != *m || islabel(*(s - 1))));
	if(!*(str = s)) break;
    }
    return 0;
#elif 1
    register char *m, *s = str;
    if(*s) {
	for( ; ; ) {
	    m = mark;  str = s--;
	    do {
		do if(!*++s) return 0;
		while(*m != *s);
	    } while(*++m);
	    while(islabel(*++s));
	    s -= m - mark;
	    if(s == str) break;
	    while(islabel(*(s - 1))) s++;
	}
	return s;
    }
    return 0;
#else
    register char *s = str;
    register char *m = mark;
    register int i;
    if(*s) for(i = 0 ; ; ) {
	if(m[i] == s[i])
	    if(m[++i]) continue;
	    else if(!islabel(s[i])) return s;
	do {
	    if(!s[i]) return 0;
	    s++;
	} while(s[i] != m[i] || islabel(*(s - 1)));
    }
    return 0;
#endif
}

/*
 * print line, scanning for marked text
 *
 * FEATURE: mark_and_print and wrap_out ought to be integrated. Right now,
 * text which is otherwise marked, is not marked when it's wrapped.
 * Marking is performed regardless of possible attribute changes in the text.
 * Cancan uses attributes, but doesn't really handle them. For instance, try
 * "#mark \ =bold", and see what output is produced for long space sequences.
 *
 * BUG: at any position in line, the longest matching marker should be used,
 * not the first one found (shouldn't depend on order of markers).
 *
 * BUG: it is possible that a letter ending an escape sequence is matched as
 * the first char of a marker.
 */
void mark_and_print(line)
char *line;
{
    static char begin[CAPLEN], end[CAPLEN];
    marknode *m, *np;

    for(m = 0, np = markers; np; np = np->next)
	if((np->pos = str_mark(line, np->pattern)))
	    if(!m || m->pos > np->pos) m = np;
    /*
     * NOTE: we should skip a str_mark if np->plen is obviousely too large.
     * or perhaps we should pass an argument to str_mark, indicating how
     * many positions to scan for match?
     */
    while(m) {
	attr_string(m->attrcode, begin, end);
	printf("%.*s%s%s%s", m->pos - line, line, begin, m->pattern, end);
	if(!*(line = m->pos + m->plen)) break;
	for(m = 0, np = markers; np; np = np->next)
	    if(np->pos)
		if(np->pos >= line || (np->pos = str_mark(line, np->pattern)))
		    if(!m || m->pos > np->pos) m = np;
    }
    fputs(line, stdout);
}

/*
 * write string to stdout, wrapping words to the next line.
 * attempt to handle the control characters.
 * return the position in the string where (last) wrap occurred, or NULL.
 *
 * FEATURE: ansi-like escape sequences assumed -- not entirely universal.
 * BUG: incomplete escape sequence at the end of the string is printed...
 */
char *wrap_out(str)
char *str;
{
    char *tail = 0;
    if(*str) {
	/* NOTE: this should always be initialized to cols_1 */
	unsigned int width = wrap ? cols_1 : (unsigned)-1;
	int cc = col0;
	char *s = str, *wp, c;
	do {
	    for(--s, wp = 0; ; ) {
		if(isprint(*++s)) {
		    /*
		     * NOTE: it might be a good idea to place a marker (trie)
		     * search here instead of calling mark_and_print...
		     */
		    if(*s == ' ') wp = s;
		    if(cc++ < width) continue;
		} else switch(*s) {
		    default: continue;
		    case '\033':if(*++s == '[') while(*++s && !isalpha(*s));
				if(*s) continue;
		    case '\0':	break;
		    case '\b':	if(--cc > 0) continue;
		    case '\r':	cc = 0;  tail = s;  continue;
		    case '\t':	wp = s;  cc |= 7;
				if(++cc < width) continue;
		}
		break;
	    }
	    if(!*s) { mark_and_print(str);  break; }
	    if(wp) s = wp;
	    c = *s;  *s = '\0';
	    mark_and_print(str);  *s = c;
	    if(wp) do s++; while(*s == ' ');
	    cc = 0;  newline();
	    tail = str = s;
	} while(*s);
	col0 = cc;
    }
    return tail;
}

/*
 * NOTE: should use a more powerful wrapper for paged_print, also for
 * subsession print...
 */
void paged_print(str)
char *str;
{
    defnode *nd;
    hilite(0);
    while(pager_text < lines_1) {
	int i;
	pager_text++;
#if 0
	str += printf("%.*s\n", cols_1, str) - 1;
	if(!*str || !wrap) return;
#else
	for(i = 0; i < cols_1 && str[i] && str[i] != '\n'; i++);
	printf("%.*s\n", i, str);
	if(!wrap || !*(str += i)) return;
	if(*str == '\n') str++;
#endif
    }
    nd = (defnode *)malloc(sizeof(defnode));
    if(!nd) syserr("malloc");
    nd->field = str_dup(str);
    if(!pager_tail)
	nd->next = 0, pager_head = nd;
    else
	nd->next = pager_tail->next, pager_tail->next = nd;
    pager_tail = nd;
}

static int pager_done()
{
    defnode *p;
    while((p = pager_head)) {
	pager_head = p->next;
	free(p->field);
	free(p);
    }
    pager_tail = 0;
    pager_text = 0;
    return 0;
}

static int pager_prompt()
{
    defnode *p;
    while((p = pager_tail = pager_head) && pager_text < lines_1) {
	paged_print(p->field);
	pager_head = p->next;
	free(p->field);  free(p);
    }
    if(p) {
	printf("%s--More--%s", modeinv, modenorm);
	return 1;
    }
    return pager_text = 0;
}

int pager(c)
char c;
{
    int n = 1;
    if(!pager_head) return pager_done();
    if(c == '\n' || (c == ' ' && (n = lines_1, 1))) pager_text -= n;
    else if(c == 'q' || (n = 0, 1)) pager_done();
    printf("%s%s", begoln, clreoln);
    if(!pager_prompt()) {
	redraw_prompt();
	if(c) redraw_input();
    }
    return n;
}

/*
 * clear the prompt if there is one
 */
void clear_prompt()
{
    if(col0) printf("%s%s", begoln, clreoln);
}

/*
 * redisplay the prompt if needed
 */
void redraw_prompt()
{
    if(pager_text && pager_prompt())
	return;
    else if(data_sent) {
	promptstr[0] = col0 = 0;
    } else if(col0) {
	hilite(0);
#if 0	/* mark_and_print() should be built into wrap_out() */
	col0 = 0; wrap_out(promptstr);
#else
	mark_and_print(promptstr);
#endif
    }
}

/*
 * see whether the window size has changed. return 1 if it has.
 */
int check_winsize()
{
    struct winsize z;
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &z) == 0 && z.ws_row && z.ws_col > 1 &&
	    (z.ws_row != lines || z.ws_col != cols)) {
	/* we assume that the bottom line is always the bottom line */
	line0 += z.ws_row - lines;
	lines = z.ws_row; lines_1 = lines - 1;
	cols = z.ws_col; cols_1 = cols - !wrapglitch;
	send_term_size(); /* NOTE: to which session are we sending it?? */
	if(line0 < 0) line0 = 0;
	return 1;
    }
    return 0;
}

/*
 * window size change signal handler:
 * we assume that the bottom line is always the bottom line
 */
void winchsig()
{
    /* we assume inputline is currently displayed */
    if(check_winsize()) {
	gotoxy(0, line0);  fputs(clreoscr, stdout);
	/* BUG: occurs when #columns becomes less than col0 */
	redraw_prompt();  redraw_input();
    }
    /* reinstall signal handler for system V unices */
    signal(SIGWINCH, winchsig);
}

